import numpy as np
from scipy.signal import correlate
from scipy.interpolate import interp1d
from itertools import combinations, combinations_with_replacement
from obspy.geodetics.base import gps2dist_azimuth
from obspy import Stream


def setup_indices(st):
    """
    ii, i1, j1 are indices:  ii points from the NxN matrix form to a vector that contains
    only the points below the diagonal while (i1,j1) points from the vector form back to the
    i,j components of the matrix.
    """
    Nseis0 = len(st)
    k = 0
    ii = np.zeros((int(Nseis0 * (Nseis0 - 1) / 2),), dtype=int)
    i1 = ii.T.copy()
    j1 = ii.T.copy()
    for j in np.arange(int(Nseis0 - 1)):
        for i in np.arange(int(j + 1), int(Nseis0)):
            ii[k] = i + j * Nseis0
            i1[k] = i
            j1[k] = j
            k = k + 1
    return ii, i1, j1


def drop_stas(st, key):
    keep = np.setdiff1d(np.arange(len(st)), key)
    ST = Stream()
    for ind in keep:
        ST += st[ind]
    return ST


def station_distances(st):

    Delta = np.zeros((len(st), len(st)))
    for comb in combinations(np.arange(len(st)), 2):
        xy1 = st[comb[0]].stats.coordinates
        xy2 = st[comb[1]].stats.coordinates
        Delta[comb[0], comb[1]] = (
            gps2dist_azimuth(xy1.latitude, xy1.longitude, xy2.latitude, xy2.longitude)[
                0
            ]
            / 1000.0
        )
    return Delta


def station_xcorr(st, mlag, minimize_type, lookup_type):
    tC = np.linspace(-mlag, mlag, 2 * mlag + 1) / st[0].stats.sampling_rate
    C = np.zeros((len(tC), int(len(st) + len(st) * (len(st) - 1) / 2)))
    LAG = np.zeros((len(st), len(st)))
    maxC = np.zeros((len(st), len(st)))

    tmp_i = list()
    tmp_j = list()
    tmp_I = np.zeros((len(st), len(st)))

    ref = correlate(st[0].data, st[0].data, mode="full", method="fft")
    mid = ref.argmax()

    for i, comb in enumerate(combinations_with_replacement(np.arange(len(st)), 2)):
        tmp_i.append(comb[0])
        tmp_j.append(comb[1])
        tmp_I[comb[0], comb[1]] = 1

        scale = np.linalg.norm(st[comb[0]].data) * np.linalg.norm(st[comb[1]].data)
        tmp = correlate(
            st[comb[0]].data, st[comb[1]].data, mode="full", method="fft"
        ) / float(scale)
        # tmp = tmp[mid-mlag-1:mid+mlag]
        tmp = tmp[
            mid - mlag : mid + mlag + 1
        ]  # Oct-30-2017 change when switching from mlag=int(2*dTmax) to mlag=int(dTmax)
        C[:, i] = tmp

        if minimize_type == "time":
            f1 = interp1d(np.linspace(0, 1, len(tmp)), tmp, kind=lookup_type)
            tmp_interp = f1(np.linspace(0, 1, 100 * len(tmp)))
            tC_interp = np.linspace(tC[0], tC[-1], len(tmp_interp))

            maxC[comb[0], comb[1]] = tmp_interp.max()
            maxC[comb[1], comb[0]] = tmp_interp.max()
            LAG[comb[0], comb[1]] = tC_interp[tmp_interp.argmax()]
            LAG[comb[1], comb[0]] = -tC_interp[tmp_interp.argmax()]
        else:
            maxC[comb[0], comb[1]] = tmp.max()
            maxC[comb[1], comb[0]] = tmp.max()
            LAG[comb[0], comb[1]] = tC[tmp.argmax()]
            LAG[comb[1], comb[0]] = -tC[tmp.argmax()]

    indx = np.array([tmp_i, tmp_j, np.where(tmp_I.flatten())[0]])

    return C, tC, indx, LAG, maxC


def trim_CC(drop_key, Nprev, Nseis0, C, maxC, LAG, ii, indx):

    key = np.setdiff1d(np.arange(Nprev), drop_key)
    keep_indx = np.zeros((int(len(key) * (len(key) - 1) / 2) + len(key),), dtype=int)

    for i, comb in enumerate(combinations_with_replacement(key, 2)):
        keep_indx[i] = comb[0] * Nprev + comb[1]

    INDX = np.zeros((3, len(keep_indx)))
    maxc = np.zeros((len(key), len(key)))
    lag = np.zeros((len(key), len(key)))
    for i, comb in enumerate(combinations_with_replacement(np.arange(len(key)), 2)):
        INDX.T[i] = [comb[0], comb[1], comb[0] * Nseis0 + comb[1]]
        maxc[comb[0]][comb[1]] = maxC[key[comb[0]]][key[comb[1]]]
        maxc[comb[1]][comb[0]] = maxC[key[comb[1]]][key[comb[0]]]
        lag[comb[0]][comb[1]] = LAG[key[comb[0]]][key[comb[1]]]
        lag[comb[1]][comb[0]] = LAG[key[comb[1]]][key[comb[0]]]

    (ind_indx,) = np.where(np.in1d(indx[2], keep_indx))
    C = C.T[ind_indx].T
    indx = INDX
    maxC = maxc
    Chat = maxC.flatten()[
        ii
    ]  # vector form of the maximum correlation coeffient for each pair
    LAG = lag

    return C, indx, LAG, maxC, Chat


def station_weights(Chat, lag, Delta, Nseis0, i1, j1, XC):
    W2 = 1 / (np.polyval(XC._p, Chat))
    (k,) = np.where(Chat < XC.Cmin)
    W2[k] = 0
    (k,) = np.where(Chat > XC._Cmax)
    W2[k] = 0
    (k,) = np.where(Delta < XC._dx_min)
    W2[k] = 0
    (k,) = np.where(np.abs(lag) > (Delta / XC._v0 + XC._dt))
    W2[k] = 0
    (k,) = np.where(Delta / XC._v0 > XC.dTmax_s)
    if len(k) > 0 and XC.output > 2:
        print(
            f"{len(k):.0f} station pairs with maximum theoretical travel times > allowed lag times"
        )
    W2[k] = 0

    staW = np.zeros(
        Nseis0,
    )
    for k in np.arange(Nseis0):
        (k1,) = np.where(i1 == k)
        (k2,) = np.where(j1 == k)
        staW[k] = np.sum(W2[k1]) / 2.0 + np.sum(W2[k2]) / 2.0

    return staW, W2


def initial_correlation(XC, st):
    drop_key = np.array([])
    count = 1
    while (drop_key > -1).any() or count == 1:
        Nseis0 = len(st)
        err = False
        ii, i1, j1 = setup_indices(st)
        Delta = station_distances(st)
        Delta = Delta.flatten()[ii]
        if not (drop_key > -1).any():
            C, tC, indx, LAG, maxC = station_xcorr(
                st, XC._mlag, XC._minimize, XC.lookup_type
            )
            Chat = maxC.flatten()[
                ii
            ]  # vector form of the maximum correlation coeffient for each pair
            count = 0
        else:
            C, indx, LAG, maxC, Chat = trim_CC(
                drop_key, Nprev, Nseis0, C, maxC, LAG, ii, indx
            )

        lag = LAG.flatten()[
            ii
        ]  # vector form of the best lag time for each station pair
        staW, W2 = station_weights(Chat, lag, Delta, Nseis0, i1, j1, XC)

        drop_key = np.where(staW == 0)[0]
        Nprev = Nseis0
        st = drop_stas(st, drop_key)

        if (
            W2.max() == 0
            or np.unique([tr.stats.station for tr in st]).size < XC.sta_min
        ):
            err = True
            drop_key = np.array([])

    if err:
        CC = dict({"err": err, "st": st})
    else:
        CC = dict(
            {
                "st": st,
                "C": C,
                "LAG": LAG,
                "maxC": maxC,
                "indx": indx,
                "Delta": Delta,
                "W2": W2,
                "staW": staW,
                "err": err,
                "ii": ii,
                "i1": i1,
                "j1": j1,
                "tC": tC,
            }
        )
    return CC


def remove_stations(CC, XC, drop_key):
    # The idea of this code is to remove selected stations and reindex behind
    # the scenes

    Nprev = len(CC["st"])
    st = drop_stas(CC["st"], drop_key)

    if not (drop_key > -1).any():
        print("No stations to remove")
        return CC

    C = CC["C"]
    maxC = CC["maxC"]
    LAG = CC["LAG"]
    indx = CC["indx"]
    while (drop_key > -1).any():
        err = False
        Nseis0 = len(st)
        ii, i1, j1 = setup_indices(st)
        Delta = station_distances(st)
        Delta = Delta.flatten()[ii]
        C, indx, LAG, maxC, Chat = trim_CC(
            drop_key, Nprev, Nseis0, C, maxC, LAG, ii, indx
        )

        lag = LAG.flatten()[
            ii
        ]  # vector form of the best lag time for each station pair
        staW, W2 = station_weights(Chat, lag, Delta, Nseis0, i1, j1, XC)

        drop_key = np.where(staW == 0)[0]
        Nprev = Nseis0
        st = drop_stas(st, drop_key)

        if (
            W2.max() == 0
            or np.unique([tr.stats.station for tr in st]).size < XC.sta_min
        ):
            err = True
            drop_key = np.array([])

    # if err:
    #     CCrm=dict({'err':err,'st':st})
    # else:
    #     CCrm=dict({'st':st,'C':C,'LAG':LAG,'maxC':maxC,
    #             'indx':indx,'Delta':Delta,'W2':W2,'staW':staW,
    #             'err':err,'ii':ii,'i1':i1,'j1':j1,'tC':CC['tC']})

    CCrm = dict(
        {
            "st": st,
            "C": C,
            "LAG": LAG,
            "maxC": maxC,
            "indx": indx,
            "Delta": Delta,
            "W2": W2,
            "staW": staW,
            "err": err,
            "ii": ii,
            "i1": i1,
            "j1": j1,
            "tC": CC["tC"],
        }
    )

    return CCrm


def bootstrap_CC(CC, XC):

    Nprev = len(CC["st"])
    st = CC["st"]

    drop_key = np.array([])
    bstrapFLAG = True

    while (drop_key > -1).any() | bstrapFLAG:
        err = False
        Nseis0 = len(st)
        if (drop_key > -1).any():
            ii, i1, j1 = setup_indices(st)
            Delta = station_distances(st)
            Delta = Delta.flatten()[ii]
            C, indx, LAG, maxC, Chat = trim_CC(
                drop_key, Nprev, Nseis0, C, maxC, LAG, ii, indx
            )
        else:
            ii = CC["ii"]
            i1 = CC["i1"]
            j1 = CC["j1"]
            Delta = CC["Delta"]
            C = CC["C"]
            LAG = CC["LAG"]
            indx = CC["indx"]
            n_data = int(np.round(XC.bootstrap_prct * len(CC["ii"])))
            a = np.floor(
                len(CC["ii"])
                * np.random.rand(
                    n_data,
                )
            ).astype("int")
            bstrap_i = ii[a]
            maxC = CC["maxC"].flatten()
            maxC[bstrap_i] = 0
            maxC = np.reshape(maxC, np.shape(LAG))
            Chat = maxC.flatten()[ii]

        lag = LAG.flatten()[
            ii
        ]  # vector form of the best lag time for each station pair
        staW, W2 = station_weights(Chat, lag, Delta, Nseis0, i1, j1, XC)

        drop_key = np.where(staW == 0)[0]
        Nprev = Nseis0
        st = drop_stas(st, drop_key)

        if (
            W2.max() == 0
            or np.unique([tr.stats.station for tr in st]).size < XC.sta_min
        ):
            err = True
            drop_key = np.array([])

        bstrapFLAG = False

    if err:
        CCnew = dict({"err": err, "st": st})
    else:
        CCnew = dict(
            {
                "st": st,
                "C": C,
                "LAG": LAG,
                "maxC": maxC,
                "indx": indx,
                "Delta": Delta,
                "W2": W2,
                "staW": staW,
                "err": err,
                "ii": ii,
                "i1": i1,
                "j1": j1,
                "tC": CC["tC"],
            }
        )

    # CCnew=dict({'st':st,'C':C,'LAG':LAG,'maxC':maxC,
    #     'indx':indx,'Delta':Delta,'W2':W2,'staW':staW,
    #     'err':err,'ii':ii,'i1':i1,'j1':j1,'tC':CC['tC']})

    return CCnew



import numpy as np
from obspy.geodetics.base import gps2dist_azimuth
from scipy.interpolate import interp1d
from scipy import ndimage

np.set_printoptions(precision=3)


def channel_list(st):
    channels = []
    for tr in st:
        tmp = ".".join(
            [tr.stats.network, tr.stats.station, tr.stats.location, tr.stats.channel]
        )
        channels.append(
            (tmp, tr.stats.coordinates.longitude, tr.stats.coordinates.latitude)
        )
    return channels


def reduced_displacement(loc, XC, st_temp):
    if np.isnan(loc.latitude):
        return np.nan

    dr = []

    channels = [ll[0] for ll in loc.channels]
    for tr in st_temp:
        if tr.id in channels:
            data = tr.data * 100
            rms = np.sqrt(np.mean(data**2))
            a = gps2dist_azimuth(
                loc.latitude,
                loc.longitude,
                tr.stats.coordinates.latitude,
                tr.stats.coordinates.longitude,
            )
            a = a[0] / 1000.0
            r = np.sqrt(a**2 + loc.depth**2) * 1000 * 100

            if "rd_freq" in XC.__dict__:
                wavelength = (XC._v0 / float(XC.rd_freq)) * 1000 * 100
                dr.append(rms * np.sqrt(wavelength * r))
            else:
                dr.append(rms * r)

    return np.median(dr)


def location_scatter(loc, lats, lons, deps):
    dist = list()
    for y, x in zip(lats, lons):
        a = gps2dist_azimuth(loc.latitude, loc.longitude, y, x)
        dist.append(a[0] / float(1000))

    scatter_x = np.median(dist)
    scatter_z = np.median(np.abs(deps - loc.depth))

    return scatter_x, scatter_z


def rotate_coords(x, y, x0, y0, angle):
    angle = np.radians(
        angle
    )  # "+" for counter-clockwise rotation , "-" for clockwise rotation
    R = np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])

    A = np.array([x, y])
    newX, newY = np.matmul(R, A)
    newX = newX + x0
    newY = newY + y0

    return newX, newY


def create_rotated_grid(rotate_params):
    import utm

    """ Example dictionary for rotate_params:

    rotate_params = { 'x'    : np.arange(0,120,8),
                      'y'    : np.arange(-15,18,3),
                      'z'    : np.arange(1,67,2),
                      'lat0' : 18.8956,
                      'lon0' : -155.6078,
                      'az'   : 120 }

    lat0,lon0 = origin
    az        = rotation angle, counterclockwise from East
    x,y,z     = grid to be rotated. +y=North, +x=East, +z=Increasing depth
    """

    # convert lat/lon origin into utm x-y origin
    x0, y0, zone_num, zone_let = utm.from_latlon(
        rotate_params["lat0"], rotate_params["lon0"]
    )

    # create blank grids
    LATS = np.ones(
        (len(rotate_params["y"]), len(rotate_params["x"]), len(rotate_params["z"]))
    )
    LONS = np.ones_like(LATS)
    DEPS = np.ones_like(LATS)

    # fill out depth grid
    for k, z in enumerate(rotate_params["z"]):
        DEPS[:, :, k] = z

    # fill out LATS and LONS grids with rotated coordinates
    for i, y in enumerate(rotate_params["y"]):
        for j, x in enumerate(rotate_params["x"]):
            newX, newY = rotate_coords(1000 * x, 1000 * y, x0, y0, rotate_params["az"])
            LATS[i, j, :], LONS[i, j, :] = utm.to_latlon(
                newX, newY, zone_number=zone_num, zone_letter=zone_let
            )

    return LONS, LATS, DEPS


def CCtoHypocenter(CC, XC):
    Nseis0 = len(CC["st"])
    NW2 = len(
        np.where(CC["W2"] > 0)[0]
    )  # number of seismogram pairs contributing to solution

    ## Calculate misfit for every location on the grid ##
    #####################################################
    dT = np.zeros(
        (len(XC._grid["LON"].flatten()), Nseis0)
    )  # delay times to each seismogram
    for ksta in np.arange(Nseis0):
        dT.T[ksta] = (
            CC["st"][ksta].TT.flatten(order="F") * CC["st"][ksta].stats.sampling_rate
        )

    if XC._minimize == "correlation":
        CCshift = np.arange(Nseis0 * (Nseis0 - 1) / 2) * (2 * (XC._mlag + 1) - 1)
        CCshift = np.tile(CCshift, len(XC._grid["LON"].flatten())).reshape(
            len(XC._grid["LON"].flatten()), len(CCshift)
        )

        deltaT = dT.T[CC["j1"]].T - dT.T[CC["i1"]].T
        deltaT[np.where(abs(deltaT) > XC._mlag)] = np.nan
        deltaT = deltaT + XC._mlag
        IND = deltaT + CCshift

        obs_corr = CC["maxC"].flatten()[CC["ii"]]
        all_corr = CC["C"].T[np.where(CC["indx"][0][:] != CC["indx"][1][:])[0]].T
        f = interp1d(
            np.arange(all_corr.size), all_corr.flatten(order="F"), kind=XC.lookup_type
        )
        pred_corr = f(IND.flatten(order="F"))
        pred_corr = np.reshape(pred_corr, IND.shape, order="F")
        residual = obs_corr - pred_corr

    elif XC._minimize == "time":
        obs_lag = CC["LAG"].flatten()[CC["ii"]]
        pred_lag = dT.T[CC["j1"]].T - dT.T[CC["i1"]].T
        residual = obs_lag - pred_lag

    if XC._normType == 1:
        residual = np.abs(residual)
    elif XC._normType == 2:
        residual = np.square(residual)

    misfit = np.einsum("...j,j", residual, CC["W2"], optimize=False) / float(NW2)
    ind_min = misfit.argmin()

    lat = XC._grid["LAT"].flatten(order="F")[ind_min]
    lon = XC._grid["LON"].flatten(order="F")[ind_min]
    dep = XC._grid["DEP"].flatten(order="F")[ind_min]

    return lat, lon, dep, misfit


def create_regrid_array(old_array, lat_inds, lon_inds, dep_inds, zoom=5, order=3):
    new_array = old_array[lat_inds, :, :]
    new_array = new_array[:, lon_inds, :]
    new_array = new_array[:, :, dep_inds]

    new_array = ndimage.zoom(new_array, zoom, order=order, mode="nearest")
    return new_array


def regridHypocenter(XC, misfit):
    misfit = np.reshape(misfit, np.shape(XC._grid["LON"]), order="F")
    inds_min = np.unravel_index(misfit.argmin(), misfit.shape)

    if XC.rotation:
        lat_slice = np.unique(
            np.clip(
                np.arange(inds_min[0] - 2, inds_min[0] + 3),
                0,
                len(XC.grid_size["y"]) - 1,
            )
        )
        lon_slice = np.unique(
            np.clip(
                np.arange(inds_min[1] - 2, inds_min[1] + 3),
                0,
                len(XC.grid_size["x"]) - 1,
            )
        )
    else:
        lat_slice = np.unique(
            np.clip(
                np.arange(inds_min[0] - 2, inds_min[0] + 3),
                0,
                len(XC.grid_size["lats"]) - 1,
            )
        )
        lon_slice = np.unique(
            np.clip(
                np.arange(inds_min[1] - 2, inds_min[1] + 3),
                0,
                len(XC.grid_size["lons"]) - 1,
            )
        )

    dep_slice = np.unique(
        np.clip(
            np.arange(inds_min[2] - 2, inds_min[2] + 3),
            0,
            len(XC.grid_size["deps"]) - 1,
        )
    )
    lons_new = create_regrid_array(
        XC._grid["LON"], lat_slice, lon_slice, dep_slice, zoom=XC.regrid, order=1
    )
    lats_new = create_regrid_array(
        XC._grid["LAT"], lat_slice, lon_slice, dep_slice, zoom=XC.regrid, order=1
    )
    deps_new = create_regrid_array(
        XC._grid["DEP"], lat_slice, lon_slice, dep_slice, zoom=XC.regrid, order=1
    )

    misfit_new = create_regrid_array(
        misfit, lat_slice, lon_slice, dep_slice, zoom=XC.regrid, order=3
    )

    new_inds_min = np.unravel_index(misfit_new.argmin(), misfit_new.shape)

    lat_new = lats_new[new_inds_min]
    lon_new = lons_new[new_inds_min]
    dep_new = deps_new[new_inds_min]

    return lat_new, lon_new, dep_new



import sys
from copy import copy, deepcopy
from itertools import combinations

import numpy as np
from obspy.geodetics import locations2degrees
from obspy.geodetics.base import gps2dist_azimuth
from obspy.taup import TauPyModel

from enveloc import xcorr_utils
from enveloc import xloc_utils


def progress(count, total, status=""):
    """
    Method to display status bar when calculating traveltimes

    Parameters
    ----------
    count : int, required
            number of current iteration
    total : int, required
            number of iterations required
    status : str, optional
            string to be printed to screen with status message

    """

    bar_len = 35
    filled_len = int(round(bar_len * count / float(total)))
    percents = round(100.0 * count / float(total), 1)
    bar = "#" * filled_len + " " * (bar_len - filled_len)
    sys.stdout.write(
        "Calculating travel times |{}| {}{} {}{}".format(
            bar, percents, "%", status, "\r"
        )
    )
    sys.stdout.flush()  # As suggested by Rom Ruben (see: http://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console/27871113#comment50529068_27871113)


def checkvalue(value):
    """
    Function to switch from 'None' to NaN

    Parameters
    ----------
    value : number, requred

    """

    if isinstance(value, float) | isinstance(value, int):
        pass
    else:
        value = np.nan
    return value


def latlon2xy(loc):
    """
    Function to change lat/lon to x/y for distance calulation in clustering method

    Parameters
    ----------
    loc : obj, required
            location object returned from :meth:`XC_locate()`

    """
    lats = np.array(loc.detections.get_lats())
    lons = np.array(loc.detections.get_lons())

    y0 = np.nanmean(lats)
    x0 = np.nanmean(lons)

    x = []
    y = []
    for lat, lon in zip(lats, lons):

        a = gps2dist_azimuth(lat, lon, y0, lon)[0] / 1000.0
        if lat < y0:
            a = -a
        y.append(a)
        a = gps2dist_azimuth(lat, lon, lat, x0)[0] / 1000.0
        if lon < x0:
            a = -a
        x.append(a)

    return x, y


def check_edgeproblems(edge_control, st_tmp):
    """
    Function to remove traces whose peak amplitude occur with 'edge_control' percent
    of the window edge

    """
    edge = np.round(edge_control * st_tmp[0].stats.npts)
    for tr in st_tmp:
        if tr.data.argmax() <= edge - 1 or tr.data.argmax() >= tr.stats.npts - edge:
            st_tmp.remove(tr)
    return st_tmp


def XC_locate(win, XC):
    """
    Main location function. Called internally from `XCOR` class's `locate()` method

    Parameters
    ----------
    win : list, required
            2 element list of obspy UTCDateTime start and end times for which to cut out a window of seismic data
    XC : obj, required
            object created by :class:`XCOR` class

    """

    st_tmp = XC.traces.slice(win[0], win[1], nearest_sample=True)

    if XC.output > 0:
        print(
            "{}  --  {}".format(
                win[0].strftime("%Y.%m.%d %H:%M:%S"),
                win[1].strftime("%Y.%m.%d %H:%M:%S"),
            )
        )

    """ Remove traces with zero-ed out data """
    for tr in st_tmp:
        if np.any(tr.data < 1e-15) and not XC._waveform_loc:
            st_tmp.remove(tr)
    if len(st_tmp) < XC.sta_min:
        if XC.output > 0:
            print("WARNING FROM XC_locate: too many traces with no data")
        return location(
            starttime=win[0], endtime=win[1], channels=xloc_utils.channel_list(st_tmp)
        )  # return latitude, longitude, depth # return because these envelopes suck.

    """ Remove traces with flagged gaps """
    for tr in st_tmp:
        if np.any(tr.data == XC._gap_value) and not XC._waveform_loc:
            print("Gap in {} | Removing station...".format(tr.id))
            st_tmp.remove(tr)
    if len(st_tmp) < XC.sta_min:
        if XC.output > 0:
            print("WARNING FROM XC_locate: too many traces with no data")
        return location(
            starttime=win[0], endtime=win[1], channels=xloc_utils.channel_list(st_tmp)
        )  # return latitude, longitude, depth # return because these envelopes suck.

    """ Remove traces with triggers within time window """
    for tr in st_tmp:
        if hasattr(tr, "triggers"):
            if np.any(
                np.logical_and(
                    tr.triggers >= tr.stats.starttime, tr.triggers <= tr.stats.endtime
                )
            ):
                st_tmp.remove(tr)
    if len(st_tmp) < XC.sta_min:
        if XC.output > 0:
            print("WARNING FROM XC_locate: too many traces with triggers")
        return location(
            starttime=win[0], endtime=win[1], channels=xloc_utils.channel_list(st_tmp)
        )  # return latitude, longitude, depth # return because these envelopes suck.

    ####################################################
    """	Check for traces with peaks at or near the edge.
		Remove problem traces and check that
		enough stations remain  """
    ####################################################
    if XC._edge_control > 0 and not XC._waveform_loc:
        st_tmp = check_edgeproblems(XC._edge_control, st_tmp)
        if np.unique([tr.stats.station for tr in st_tmp]).size < XC.sta_min:
            if XC.output > 0:
                print(
                    "WARNING FROM XC_locate: too many traces with peaks at edge of window"
                )
            return location(
                starttime=win[0],
                endtime=win[1],
                channels=xloc_utils.channel_list(st_tmp),
            )  # return latitude, longitude, depth # return because these envelopes suck.
    ####################################################
    ####################################################

    ####################################################
    """ perform initial cross-correlation of all traces """
    if XC.detrend:
        st_tmp.detrend("demean")
    CC = xcorr_utils.initial_correlation(XC, st_tmp)
    """ return if not enough stations remain """
    if CC["err"]:
        if XC.output > 1:
            print(
                "WARNING FROM XC_locate: need at least {:.0f} stations to correlate above {:.2f}".format(
                    XC.sta_min, XC.Cmin
                )
            )
        return location(
            starttime=win[0], endtime=win[1], channels=xloc_utils.channel_list(CC["st"])
        )  # return latitude, longitude, depth # return because these envelopes suck.
    ####################################################
    ####################################################

    """ make working copy of cross-correlation dictionary """
    CCrm = CC.copy()

    count = 0
    drop_key = np.array([])
    """ you're going stay in the loop if it is the first time (count==0) or 
	    while there are stations to remove and the data need relocated from
	    interactive mode """
    while (drop_key > -1).any() or count == 0:

        bstrap_lat = list()
        bstrap_lon = list()
        bstrap_dep = list()

        """ remove stations and values from cc-matrix if stations
		    get removed during interactive location process   """
        if (drop_key > -1).any():
            if XC.output > 1:
                print("Removing stations.")
            CCrm = xcorr_utils.remove_stations(CCrm, XC, drop_key)

        nan_count = 0
        """ Loop for bootstrap iterations. Last iteration is uses all of the data. """
        for bstrap in np.arange(XC.bootstrap):

            ####################################################
            """Check if you've had too many bad (NaN) bootstrap locations"""
            if nan_count >= XC.bootstrap / 2.0:
                if XC.output > 1:
                    print("Too many NaN bootstrap locations.")
                return location(
                    starttime=win[0],
                    endtime=win[1],
                    channels=xloc_utils.channel_list(CCrm["st"]),
                )
            ####################################################
            ####################################################

            ####################################################
            """ make boostrap working copy of cross-correlation dictionary """
            CCnew = CCrm.copy()
            """ If not the last iteration, throw away % of cc-matrix data  """
            if bstrap != np.arange(XC.bootstrap)[-1]:
                CCnew = xcorr_utils.bootstrap_CC(CCrm, XC)
            """ Not enough stations remain? """
            if CCnew["err"]:
                if XC.output > 1:
                    print(
                        "WARNING FROM XC_locate: need at least {:.0f} stations to correlate above {:.2f}".format(
                            XC.sta_min, XC.Cmin
                        )
                    )
                """ Continue if bootstrap loop and iterate # of bad bstrap locations """
                if bstrap != np.arange(XC.bootstrap)[-1]:
                    nan_count = nan_count + 1
                    continue
                ### Return empty location object if it is last iteration   """
                else:
                    return location(
                        starttime=win[0],
                        endtime=win[1],
                        channels=xloc_utils.channel_list(CCrm["st"]),
                    )  # return latitude, longitude, depth # return because these envelopes suck.
            ####################################################
            ####################################################

            """ Do a grid search to get a hypocenter """
            tmp_lat, tmp_lon, tmp_dep, misfit = xloc_utils.CCtoHypocenter(CCnew, XC)
            if "windows" not in XC.__dict__ and XC._num_processors == 1:
                XC.misfit = misfit

            ####################################################
            """ Check if location is on grid edge """
            edge_check = False
            if XC.rotation:
                if (
                    np.in1d(tmp_lat, XC._grid["LAT"][0, :, 0])[0]
                    or np.in1d(tmp_lat, XC._grid["LAT"][-1, :, 0])[0]
                ):
                    edge_check = True
                if (
                    np.in1d(tmp_lat, XC._grid["LON"][0, :, 0])[0]
                    or np.in1d(tmp_lat, XC._grid["LON"][-1, :, 0])[0]
                ):
                    edge_check = True
            else:
                if (
                    np.in1d(tmp_lat, XC.grid_size["lats"].take([0, -1]))[0]
                    or np.in1d(tmp_lon, XC.grid_size["lons"].take([0, -1]))[0]
                ):
                    edge_check = True
            if edge_check:
                if XC.output > 1:
                    print("On grid edge!! No location for you.")
                if bstrap != np.arange(XC.bootstrap)[-1]:
                    nan_count = nan_count + 1
                    continue
                else:
                    return location(
                        starttime=win[0],
                        endtime=win[1],
                        channels=xloc_utils.channel_list(CCrm["st"]),
                    )
            ####################################################
            ####################################################

            """ Regrid if XC.regrid==True """
            if XC.regrid:
                tmp_lat, tmp_lon, tmp_dep = xloc_utils.regridHypocenter(XC, misfit)

            ####################################################
            """ Append bootstrap location info if not last iteration """
            if bstrap != np.arange(XC.bootstrap)[-1]:
                if XC.output > 1:
                    print(
                        "...bootstrap...Latitude: {:.3f}, Longitude: {:.3f}, Depth {:.1f}".format(
                            tmp_lat, tmp_lon, tmp_dep
                        )
                    )
                bstrap_lat.append(tmp_lat)
                bstrap_lon.append(tmp_lon)
                bstrap_dep.append(tmp_dep)
            ### Last iteration. Set location object info with this location. """
            else:
                loc = location(
                    latitude=tmp_lat,
                    longitude=tmp_lon,
                    depth=tmp_dep,
                    starttime=win[0],
                    endtime=win[1],
                    channels=xloc_utils.channel_list(CCrm["st"]),
                )
                """ Optionally calculate reduced displacement"""
                if "raw_traces" in XC.__dict__:
                    loc.reduced_displacement = xloc_utils.reduced_displacement(
                        loc, XC, XC.raw_traces.slice(loc.starttime, loc.endtime)
                    )
                """ If bootstraping at all, calculate scatter from bootstrap locations """
                if XC.bootstrap > 1:
                    dx, dz = xloc_utils.location_scatter(
                        loc, bstrap_lat, bstrap_lon, bstrap_dep
                    )
                    loc.horizontal_scatter = dx
                    loc.vertical_scatter = dz
                    if XC.output > 0:
                        print(
                            "--Location--\nLatitude: {:.3f}, Longitude: {:.3f}\t +/- {:.1f}\n--Depth--\n{:.1f} +/- {:.1f} km".format(
                                tmp_lat, tmp_lon, dx, tmp_dep, dz
                            )
                        )
                else:
                    if XC.output > 0:
                        print(
                            "--Location--\nLatitude: {:.3f}, Longitude: {:.3f}, Depth: {:.1f} km".format(
                                tmp_lat, tmp_lon, tmp_dep
                            )
                        )
            ####################################################
            ####################################################

        ####################################################
        """ Done with bootstrap loop. Plot & interact """
        if XC.plot and XC._num_processors == 1:
            from enveloc.plotting_utils import XC_plot

            """ set up some variables for plotting """
            CC1 = 1 - CCrm["C"][:, CCrm["indx"][0, :] != CCrm["indx"][1, :]]
            Nseis0 = len(CCrm["st"])
            plot_loc = loc.copy()
            plot_loc.bstrap_lat = np.array(bstrap_lat)
            plot_loc.bstrap_lon = np.array(bstrap_lon)
            CCshift = (
                np.arange(Nseis0 * (Nseis0 - 1) / 2) * (2 * (XC._mlag + 1) - 1) - 1
            )  # index into cross-corrleation array
            """ Call plotting routine. Return plot_opt variable to see if 
			    we need to relocate """
            plot_opt = XC_plot(CCrm, XC, CC1, misfit, plot_loc)
            drop_key = plot_opt["krm"]
            """ relocate with some stations removed """
            if plot_opt["relocate"]:
                count = -1
            elif plot_opt["restart"]:
                """relocate with original station list"""
                count = -1
                CCrm = CC.copy()
                drop_key = np.array([])
        ####################################################
        ####################################################

        count = count + 1
    """ done with while loop looking for revomved station """

    return loc


class location(object):
    """Define a location object"""

    def __init__(
        self,
        latitude=None,
        longitude=None,
        depth=None,
        dx=None,
        dz=None,
        starttime=None,
        endtime=None,
        channels=[],
        reduced_displacement=None,
    ):
        self.latitude = checkvalue(latitude)
        self.longitude = checkvalue(longitude)
        self.depth = checkvalue(depth)
        self.horizontal_scatter = checkvalue(dx)
        self.vertical_scatter = checkvalue(dz)
        self.starttime = starttime
        self.endtime = endtime
        self.channels = channels
        self.reduced_displacement = reduced_displacement

    def __repr__(self):
        a = self.__dict__.keys()
        A = []
        for item in a:
            A.append(item + "=" + str(self.__getattribute__(item)))
        return "location(" + "\n".join(A) + ")"

    def station_latlons(self):
        """return stas, np.array(lats), np.array(lons) for this location"""
        lats = []
        lons = []
        stas = []
        for c in self.channels:
            lats.append(c[2])
            lons.append(c[1])
            stas.append(c[0])

        return stas, np.array(lats), np.array(lons)

    def copy(self):
        return deepcopy(self)


class event_list(object):
    """
    Define an events object, which is a list of :class:`~enveloc.core.location` objects

    """

    def __init__(self, events=[]):
        self.events = events

    def __repr__(self):
        return "event_list object containing {:.0f} events".format(len(self.events))

    def get_lats(self):
        """
        Return numpy array of all location latitudes in the list

        """
        lats = np.array([a.latitude for a in self.events])
        return lats

    def get_lons(self):
        """
        Return numpy array of all location longitudes in the list

        """
        lons = np.array([a.longitude for a in self.events])
        return lons

    def get_depths(self):
        """
        Return numpy array of all location depths in the list

        """
        lons = np.array([a.depth for a in self.events])
        return lons

    def get_times(self):
        """
        Return two numpy arrays of all location window start and end times in the list

        """
        starttimes = np.array([a.starttime for a in self.events])
        endtimes = np.array([a.endtime for a in self.events])
        return starttimes, endtimes

    def get_reduced_displacement(self):
        """
        Return all location reduced displacements in the events list as a numpy array

        """
        reduced_displacement = np.array([a.reduced_displacement for a in self.events])
        return reduced_displacement

    def get_numchannels(self):
        """
        Returns a numpy array of number of channels used in each location

        """
        num = np.array([len(a.station_latlons()[0]) for a in self.events])
        return num

    def get_stations(self):
        """
        Return a list of all stations used in all locations in event list

        """
        stas = []
        for a in self.events:
            for sta in a.channels:
                stas.append(sta[0])
        return stas

    def tolist(self):
        """
        returns the attribute `events` as a list

        """
        return self.events

    def filter(
        self,
        min_t=None,
        max_t=None,
        min_lat=None,
        max_lat=None,
        min_lon=None,
        max_lon=None,
        min_horizontal_scatter=None,
        max_horizontal_scatter=None,
        min_depth=None,
        max_depth=None,
        min_rd=None,
        max_rd=None,
        min_vertical_scatter=None,
        max_vertical_scatter=None,
        min_num_channels=None,
        max_num_channels=None,
        highpass_loc=None,
        inplace=False,
    ):
        """
        Filter this event list by location properties and return a new event list object

        Parameters
        ----------
        Mostly self-explanatory

        """

        NEW = self.copy()
        NEW.events = np.array(NEW.events)
        if min_t:
            NEW.events = NEW.events[NEW.get_times()[0] > min_t]
        if max_t:
            NEW.events = NEW.events[NEW.get_times()[0] < max_t]
        if min_lat:
            NEW.events = NEW.events[NEW.get_lats() > min_lat]
        if max_lat:
            NEW.events = NEW.events[NEW.get_lats() < max_lat]
        if min_lon:
            NEW.events = NEW.events[NEW.get_lons() > min_lon]
        if max_lon:
            NEW.events = NEW.events[NEW.get_lons() < max_lon]
        if min_horizontal_scatter:
            d_xy = np.array([a.horizontal_scatter for a in self.events])
            NEW.events = NEW.events[d_xy > min_horizontal_scatter]
        if max_horizontal_scatter:
            d_xy = np.array([a.horizontal_scatter for a in self.events])
            NEW.events = NEW.events[d_xy < max_horizontal_scatter]
        if min_depth:
            NEW.events = NEW.events[NEW.get_depths() > min_depth]
        if max_depth:
            NEW.events = NEW.events[NEW.get_depths() < max_depth]
        if min_rd:
            NEW.events = NEW.events[NEW.get_reduced_displacement() > min_rd]
        if max_rd:
            NEW.events = NEW.events[NEW.get_reduced_displacement() < max_rd]
        if min_vertical_scatter:
            d_xy = np.array([a.vertical_scatter for a in self.events])
            NEW.events = NEW.events[d_xy > min_vertical_scatter]
        if max_vertical_scatter:
            d_xy = np.array([a.vertical_scatter for a in self.events])
            NEW.events = NEW.events[d_xy < max_vertical_scatter]
        if min_num_channels:
            num = np.array([len(a.channels) for a in self.events])
            NEW.events = NEW.events[num > min_num_channels]
        if max_num_channels:
            num = np.array([len(a.channels) for a in self.events])
            NEW.events = NEW.events[num < max_num_channels]
        if highpass_loc is not None:
            vals = np.array([l.highpass_loc for l in self.events])
            if highpass_loc:
                NEW.events = NEW.events[vals]
            else:
                NEW.events = NEW.events[~vals]
        if inplace:
            setattr(self, "events", NEW.__getattribute__("events"))
        else:
            return NEW
            # return event_list(NEW.events.tolist())

    def calc_reduced_displacement(self, st, XC, num_processors=1):
        """
        Method to calculate a reduced displacement for all locations in the events list.
        The method updates the `reduced_displacement` property in each location object. :class:`XCOR` can also do
        this internally while generating a location, but the idea with this method is to apply the calculation
        after obtaining all initial locations and throwing out bad locations via clustering or maximum horizontal
        scatter. That way you are not spending computation time for events you will never use.

        Parameters
        ----------

        st : stream, required
                Obspy stream of envelopes. Each trace must have stats.coordinates.latitude/longitude
        XC : `XCOR` object, required
                :class:`XCOR` object used to obtain locations within the event list
        num_processors : int, optional
                Optional integer to parallelize when set >1, but this doesn't appear to be faster. Default = 1.

        """
        if num_processors > 1:
            from multiprocessing import Pool

            pool = Pool(processes=num_processors)
            new_windows = []
            for l in self.events:
                if not np.isnan(l.latitude):
                    new_windows.append(l)
            results = [
                pool.apply_async(
                    xloc_utils.reduced_displacement,
                    args=(win, XC, st.slice(win.starttime, win.endtime)),
                )
                for win in new_windows
            ]
            pool.close()
            pool.join()
            RD = [p.get() for p in results]
            times = np.array(self.get_times()[0])
            for i, rd in enumerate(RD):
                ind = np.where(times == new_windows[i].starttime)[0][0]
                self.events[ind].reduced_displacement = rd
        else:
            for l in self.events:
                if not np.isnan(l.latitude):
                    l.reduced_displacement = xloc_utils.reduced_displacement(
                        l, XC, st.slice(l.starttime, l.endtime)
                    )

    def remove(self, max_scatter=3.0, rm_nan_loc=True, rm_nan_err=True, inplace=False):
        """
        Method to remove locations from the event list. Returns a new copy.

        Parameters
        ----------

        max_scatter : float, optional
                maximum horizontal scatter allowed for a give locationDefault = 3.0
        rm_nan_loc : bool, optional
                `True` will remove all locations with `nan` locations. Default = `True`
        rm_nan_err : bool, optional
                  `True` will remove all locations with nan horizontal_scatter (not appropriate
                  if not bootstrapping to obtain horizontal_scatter, eg. bootstrap=1), Default = `True`
        inplace : bool, optional
                `True` to operate in place. `False` to return a copy. Default = `False`

        """

        errs = np.array([a.horizontal_scatter for a in self.events])
        IND = np.arange(len(errs)).tolist()
        if rm_nan_loc:
            lats = self.get_lats()
            for ind in np.arange(len(errs)):
                if np.isnan(lats[ind]):
                    IND.remove(ind)

        if rm_nan_err:
            for ind in np.arange(len(errs)):
                if np.isnan(errs[ind]):
                    if ind in IND:
                        IND.remove(ind)

        for ind in np.arange(len(errs)):
            if np.isfinite(errs[ind]) and errs[ind] > max_scatter:
                if ind in IND:
                    IND.remove(ind)

        events = [self.events[ind] for ind in IND]
        if inplace:
            self.events = events
        else:
            return event_list(events)

    def remove_highpass(self, inplace=False):
        """
        Method to remove locations that also have a location with a highpass filter

        Parameters
        ----------

        inplace : bool, optional
                `True` to operate in place or `False` to return a copy. Default = `False`

        """

        NEW = self.copy()
        NEW.events = np.array(NEW.events)
        vals = np.array([l.highpass_loc for l in self.events])
        NEW.events = NEW.events[~vals]
        if inplace:
            setattr(self, "events", NEW.__getattribute__("events"))
        else:
            return NEW

    def remove_duplicates(self, distance=25.0, inplace=False):
        """
        Method to remove locations within a list that have identical starttimes

        Parameters
        ----------

        distance : float, optional
                maximum horizontal distance (km) below which the locations of events
                with identical starttimes are averaged. Default = 25.0.
        inplace : bool, optional
                `True` to operate in place or `False` to return a copy. Default = `False`

        """
        T = self.get_times()[0]
        idx_sort = np.argsort(T)
        sorted_T_array = T[idx_sort]
        vals, idx_start, count = np.unique(
            sorted_T_array, return_counts=True, return_index=True
        )
        res = np.split(idx_sort, idx_start[1:])
        vals = vals[count > 1]
        res = filter(lambda x: x.size > 1, res)

        IND = np.arange(len(T)).tolist()
        for r in res:
            for comb in combinations(r, 2):
                xy1 = self.events[comb[0]]
                xy2 = self.events[comb[1]]
                dist = (
                    gps2dist_azimuth(
                        xy1.latitude, xy1.longitude, xy2.latitude, xy2.longitude
                    )[0]
                    / 1000.0
                )
                if dist < distance:
                    if len(self.events[comb[0]].channels) > len(
                        self.events[comb[1]].channels
                    ):
                        ind = comb[0]
                    elif len(self.events[comb[0]].channels) < len(
                        self.events[comb[1]].channels
                    ):
                        ind = comb[1]
                    elif len(self.events[comb[0]].channels) == len(
                        self.events[comb[1]].channels
                    ):
                        if np.isnan(
                            self.events[comb[0]].horizontal_scatter
                        ) and not np.isnan(self.events[comb[1]].horizontal_scatter):
                            ind = comb[1]
                        elif not np.isnan(
                            self.events[comb[0]].horizontal_scatter
                        ) and np.isnan(self.events[comb[1]].horizontal_scatter):
                            ind = comb[0]
                        elif np.isnan(
                            self.events[comb[0]].horizontal_scatter
                        ) and np.isnan(self.events[comb[1]].horizontal_scatter):
                            ind = comb[0]
                        else:
                            scatter = np.array(
                                [
                                    self.events[comb[0]].horizontal_scatter,
                                    self.events[comb[1]].horizontal_scatter,
                                ]
                            )
                            ind = comb[scatter.argmin()]
                    if ind in IND:
                        IND.remove(ind)
        events = [self.events[ind] for ind in IND]
        if inplace:
            self.events = events
        else:
            return event_list(events)

    def cluster(self, dx=10, dt=60, num_events=4):
        """
        Cluster locations in the event_list using :meth:`sklearn.cluster.DBSCAN`

        Parameters
        ----------

        dx : float, optional
                Maximum horizontal distance in km. Default = 10.0
        dt : float, optional
                Maximum time difference between detections, in minutes. Default = 60
        num_events : int, optional
                Number of events required within `dx` and `dt`. Default = 4.


        The method returns a :class:`core.detections` object, which contains various different lists as attributes:

        * detections - events from original event list
        * core_clustered - events who all meet the criteria
        * edge_clustered - events within `dx` & `dt` distance of core_clustered' event, but don't themselves have `num_events` within `dx` & `dt` of them
        * noise          - events that don't meet either criteria above
        * all_clustered  - core_clustered + edge_clustered combined for convenience

        """
        det = detections(self.events)
        det.cluster(dx=dx, dt=dt, num_events=num_events)
        return det

    def plot_locations(self, XC):
        """
        Method to plot the locations for this `event_list`

        Parameters
        ----------

        XC : `XCOR` object, required
                :class:`XCOR` object used to create `event_list` object

        """

        from enveloc.plotting_utils import plot_locations

        plot_locations(self, XC)

        return

    def __add__(self, other):
        return event_list(self.events + other.events)

    def copy(self):
        """Return a copy of the events object"""
        return deepcopy(self)


class detections(object):
    """
    `detections` object. Contains different :class:`~enveloc.core.event_list` objects as properties. This
    object allows for lists all events, clustered events, and unclustered events to
    exist all in one place and be modified using the same class methods.

    """

    def __init__(self, detects=[]):
        if detects.__class__.__name__ == "list":
            self.detections = event_list(detects)
        elif detects.__class__.__name__ == "event_list":
            self.detections = detects

    def __repr__(self):
        a = self.__dict__.keys()
        A = []
        for item in a:
            A.append(
                item + ": " + str(len(self.__getattribute__(item).events)) + " events"
            )
        return "DETECTION object with attributes:\n(" + "\n".join(A) + ")"

    def copy(self):
        return deepcopy(self)

    def filter(
        self,
        min_t=None,
        max_t=None,
        min_lat=None,
        max_lat=None,
        min_lon=None,
        max_lon=None,
        min_horizontal_scatter=None,
        max_horizontal_scatter=None,
        min_depth=None,
        max_depth=None,
        min_rd=None,
        max_rd=None,
        min_vertical_scatter=None,
        max_vertical_scatter=None,
        min_num_channels=None,
        max_num_channels=None,
    ):
        """
        Filter each event_list in the detections object by location properties and return a new detections object
        with modified event_lists

        Parameters
        ----------
        Mostly self-explanatory

        """
        NEW = self.copy()
        a = NEW.__dict__.keys()
        for item in a:
            setattr(
                NEW,
                item,
                NEW.__getattribute__(item).filter(
                    min_t=min_t,
                    max_t=max_t,
                    min_lat=min_lat,
                    max_lat=max_lat,
                    min_lon=min_lon,
                    max_lon=max_lon,
                    min_horizontal_scatter=min_horizontal_scatter,
                    max_horizontal_scatter=max_horizontal_scatter,
                    min_depth=min_depth,
                    max_depth=max_depth,
                    min_rd=min_rd,
                    max_rd=max_rd,
                    min_vertical_scatter=min_vertical_scatter,
                    max_vertical_scatter=max_vertical_scatter,
                    min_num_channels=min_num_channels,
                    max_num_channels=max_num_channels,
                ),
            )
        return NEW

    def cluster(self, dx=25, dt=None, num_events=4):
        """
        Cluster locations in the 'detections' event_list using :meth:`sklearn.cluster.DBSCAN`

        Parameters
        ----------

        dx : float, optional
                Maximum horizontal distance in km. Default = 25.0
        dt : float, optional
                Maximum time difference between detections, in minutes. Default = `None`
        num_events : int, optional
                Number of events required within `dx` and `dt`. Default = 4.


        The method leaves the 'detections' `event_list` untouched, but adds additional
        event_list properties in place:

        * core_clustered - events who all meet the criteria
        * edge_clustered - events within `dx` & `dt` distance of core_clustered' event, but don't themselves have `num_events` within `dx` & `dt` of them
        * noise          - events that don't meet either criteria above
        * all_clustered  - core_clustered + edge_clustered combined for convenience

        """

        from sklearn.cluster import DBSCAN

        x, y = latlon2xy(self)

        if not dt:
            X = np.array([x, y]).T
        else:
            # scale time to match distance
            t = self.detections.get_times()[0]
            dtime = np.array(
                [(t0.datetime - t.min().datetime).total_seconds() / 60.0 for t0 in t]
            )
            dtime = dtime * (dx / float(dt))
            # put distance and time together
            X = np.array([x, y, dtime]).T

        db = DBSCAN(eps=dx, min_samples=num_events).fit(X)

        all_detects = np.where(db.labels_ > -1)[0]
        tmp = []
        for ind in all_detects:
            tmp.append(self.detections.events[ind])
        self.all_clustered = event_list(tmp)

        tmp = []
        for ind in db.core_sample_indices_:
            tmp.append(self.detections.events[ind])
        self.core_clustered = event_list(tmp)

        tmp = []
        for ind in all_detects:
            if ind not in db.core_sample_indices_:
                tmp.append(self.detections.events[ind])
        self.edge_clustered = event_list(tmp)

        tmp = []
        noise_inds = np.where(db.labels_ == -1)[0]
        for ind in noise_inds:
            tmp.append(self.detections.events[ind])
        self.noise = event_list(tmp)

    def uncluster(self):
        """
        Method to remove all event lists not named 'detections'. Operates in place.

        """
        a = self.__dict__.keys()
        for item in a:
            if item != "detections":
                self.__delattr__(item)

    def remove(self, max_scatter=3.0, rm_nan_loc=True, rm_nan_err=True, inplace=False):
        """
        Method to remove locations from the 'detections' event list. Returns a new copy.

        Parameters
        ----------

        max_scatter : float, optional
                maximum horizontal scatter allowed for a give locationDefault = 3.0
        rm_nan_loc : bool, optional
                `True` will remove all locations with `nan` locations. Default = `True`
        rm_nan_err : bool, optional
                  `True` will remove all locations with nan horizontal_scatter (not appropriate
                  if not bootstrapping to obtain horizontal_scatter, eg. bootstrap=1), Default = `True`
        inplace : bool, optional
                `True` to operate in place. `False` to return a copy. Default = `False`

        """

        NEW = self.copy()
        for item in NEW.__dict__.keys():
            setattr(
                NEW,
                item,
                NEW.__getattribute__(item).remove(
                    max_scatter=max_scatter,
                    rm_nan_loc=rm_nan_loc,
                    rm_nan_err=rm_nan_err,
                ),
            )
        if inplace:
            for item in self.__dict__.keys():
                setattr(self, item, NEW.__getattribute__(item))
        else:
            return NEW

    def remove_duplicates(self, distance=25.0, inplace=False):
        """
        Method to remove locations within a list that have identical starttimes

        Parameters
        ----------

        distance : float, optional
                maximum horizontal distance (km) below which the locations of events with
                identical starttimes are averaged. Default = 25.0.
        inplace : bool, optional
                `True` to operate in place or `False` to return a copy. Default = `False`

        """

        NEW = self.copy()
        for item in NEW.__dict__.keys():
            setattr(
                NEW,
                item,
                NEW.__getattribute__(item).remove_duplicates(distance=distance),
            )
        if inplace:
            for item in self.__dict__.keys():
                setattr(self, item, NEW.__getattribute__(item))
        else:
            return NEW

    def plot_locations(self, XC):
        """
        Method to plot the locations from each `event_list`

        Parameters
        ----------

        XC : `XCOR` object, required
                :class:`XCOR` object used to create `detections` object

        """

        from enveloc.plotting_utils import plot_locations

        plot_locations(self, XC)

        return

    def __add__(self, other):
        NEW = self.copy()
        for item in NEW.__dict__.keys():
            try:
                setattr(
                    NEW, item, NEW.__getattribute__(item) + other.__getattribute__(item)
                )
            except:
                print("Cannot add unlike objects")
                return
        return NEW


class XCOR(object):
    """
    :class:`XCOR` is the main class object that needs to be set up in order to get a location.
    Most of the paramaters will self-assign, and in many cases that is fine. The only
    essential variable is the obspy stream ``st``, but you will likely want to give thought
    and care to the grid and velocity model.

    Parameters
    ----------

    st : stream, required
            Obspy stream of envelopes. Each trace must have stats.coordinates.latitude/longitude
    model : str, optional
            Obspy taup model or model file from which to calculate travel times.
            If this is not provided, a default file will be called.
    model_dir : str, optional
            Directory where model.npz file is place by obspy's taup program
            If not provided, it will default to within the enveloc module directory
    grid_size : dict, optional
            A dict with keys ``lats``, ``lons``, and ``deps`` each of which are 1D monotonic numpy arrays.
            If unprovided, this will be created internally based on the extent of the input stations.
    detrend : bool, optional
            True/False to force a demean on obspy stream ``st`` or not.
            Default = `True`
    regrid : bool, optional
            Whether to reinterpolate location on finer grid.
            Default = `True`
    phase_types : list, optional
            list of phases for which to calculate travel times. Ultimate travel time
            used for each grid node for each station will be the minimum calculated
            from the phases in the list provided. Also accepts ['Nkmps'] where N is a
            constant velocity in km/s. This would be useful for inputing a surface wave
            velocity, for example. See obspy taup documentation for more details on phase
            nomenclature.
            Default = [ 's', 'S' ]
    minimize_type : str, optional
            'correlation' to minimize the difference between the observed Cmax and predicted C
            'time' to minimize the difference between the observed and predicted dt
            Default = 'correlation
    normType : int, optional
            Integer 1 or 2 to use an L1 or L2 norm.
            Default = 1
    plot : bool, optional
            Boolean flag to plot or not.
            Default = `True`
    interact : bool, optional
            Boolean flag to interact with plot or not.
            Default = `True`
    output : int, optional
            Interger from 0-3 controlling increasing level of messages the code produces
            Default = 1
    Cmin : float, optional
            Minimum normalized cross-correlation coefficient to be considered.
            Default = 0.5
    Cmax : float, optional
            Maximum normalized cross-correlation coefficient to be considered. Not always
            necessary, but could be for removing correlated noise spikes, for example.
            Default = 0.995
    sta_min : int, optional
            Minimum number of stations needed to obtain a location.
            Default = 3
    dx_min : float, optional
            Minimum horizontal distance (km) between channels required to consider th
            cross-correlation. Useful to exclude correlations between multiple channels
            at the same station.
            Default = 0.1
    bootstrap : int, optional
            Number of iterations for each location. For each iteration, ``bootstrap_prct``
            of the CC values will be zeroed out, and the resulting station contributions
            are adjusted accordingly to obtain a new location. These iterated locations
            provide a measure of location scatter, which is recorded in the location object.
            The final location will use all of the data (no CC values zeroed out). Thus,
            if boootstrap=N, it will bootstrap N times and get a location using all the data
            on the N+1 iteration. If bootstrap==1, the data are only located once with all the
            data, and no horizontal scatter is estimated.
            Default = 1
    bootstrap_prct : float, optional
            Value between 0 and 1 determining the fraction of cc data to throw out in each
            bootstrap iteration.
            Default = 0.04 (4%)
    lookup_type : str, optional
            Type of interpolation method to use when getting a predicted cross-correlation
            value for each channel pair for each grid ('linear', 'nearest', 'zero', 'slinear',
            'quadratic', 'cubic'). See scipy.interpolate.interp1d for details.
            Default='cubic'
    rotation : dict, optinal
            Dictionary defining desired rotated grid. Needs keys ``x`` (x grid nodes),
            ``y`` (y grid nodes), ``z`` (depth grid nodes), ``lat0`` & ``lon0`` (origin lat/lon),
            ``az``, rotation azimuth (counterclockwise from East)
            Default = ``None``
    dTmax_s : float, optional
            Maximum cross-correlation shift in seconds. Defaults to the smaller of

            a) 1/2 of the obspy stream window length

            b) the maximum predicted inter-station differential time + `dt`
    dt : float, optional
            Seconds of additional allowed cross-correlation shifts beyond the maximumpredicted
            interchannel differential time.
            Default = 3.0
    rd_freq	: float, optional
            Frequency at which to calculate surface wave reduced displacement using the
            velocity from the top layer of the velocity model. If not set, body wave reduced
            displacement will be calculated.
            Default = `None`
    raw_traces : stream, optional
            Obpsy stream of non-envelope traces matching the input ``st`` variable. Used to
            calculate reduced displacement.
            Default = `[]`
    env_hp : stream, optional
            Obspy stream of second envelopes with traces matchin the input ``st`` variable.
            This would typically be envelopes generated in a different passband than ``st``.
            If a location is obtained for ``st``, it will also check env_hp and set a flag.
            This can be useful for weeding out earthquake detections or correlated noise
            spikes when you are only interested in detecting a band-limited signal. It
            essentially gives you a way to flag windows where you have coherent energy in
            two different passbands. Kind of a niche-use case, but necessary for my needs.
            Default = `[]`
    edge_control : float, optional
            Value between 0 and 1 defining a percentage from the window edge to not allow
            the peak trace energy. That is, if any trace's peak amplitude occurs between a
            with a percentage of the window edge, it is removed. Useful when autolocating.
            Default = 0.03
    num_processors : int, optional
            Number of processors to use if you are iterating through sliding windows to
            get locations. This can also be set in the :meth:`locate()` method.
            Default = 1.
    tt_file : str, optional
            Path to a .npz file containing pre-calculated traveltimes for this station set
            on this grid. Calulated using :meth:`save_traveltimes()` method.
            Default = `None`
    waveform_loc : bool, optional
            Boolean flag to optionally locate waveforms rather than envelopes.
            This flag exists to bypass some of the built in envelope quailty control
            Default = `False`
    gap_value : int, optional
            Value used to flag data gaps identified in preprocessing.
            Default = -123454321

    return : object

    """

    def __init__(
        self,
        st,
        model=None,
        grid_size=None,
        detrend=True,
        regrid=False,
        phase_types=["s", "S"],
        normType=1,
        plot=True,
        interact=True,
        output=1,
        minimize_type="correlation",
        Cmin=0.5,
        Cmax=0.995,
        sta_min=3,
        dx_min=0.1,
        dt=3.0,
        bootstrap=1,
        bootstrap_prct=0.04,
        lookup_type="cubic",
        rotation=None,
        dTmax_s=None,
        rd_freq=None,
        raw_traces=[],
        env_hp=[],
        edge_control=0.03,
        num_processors=1,
        tt_file=None,
        waveform_loc=False,
        model_dir=None,
        gap_value=-123454321,
    ):

        if output == True:
            output = 1
        elif output == False:
            output = 0
        self.output = output
        self.traces = st
        self.detrend = detrend
        self.rotation = rotation
        if not tt_file:
            import os

            if not model_dir:
                self._model_dir = os.path.dirname(os.path.realpath(__file__))
            else:
                self._model_dir = model_dir
            if not model:
                print("Warning!  No model provided.")
                print("Default model used. Maybe not ideal...")
                self._model_file = (
                    os.path.dirname(os.path.realpath(__file__))
                    + "/default_vel_model.tvel"
                )
                print("Using " + self._model_file)
            else:
                self._model_file = model

            if type(phase_types) == str:
                phase_types = [phase_types]
            self.phase_types = phase_types

            if grid_size:
                self.grid_size = grid_size
            else:
                if self.rotation:
                    self.grid_size = dict(
                        {
                            "x": self.rotation["x"],
                            "y": self.rotation["y"],
                            "deps": self.rotation["z"],
                        }
                    )
                else:
                    self.grid_size = dict({"lats": [], "lons": [], "deps": []})

                    print("Warning!  No grid provided.")
                    print("Making grid based on stations provided...")

                    lats = np.array([tr.stats.coordinates.latitude for tr in st])
                    lons = np.array([tr.stats.coordinates.longitude for tr in st])
                    dlat = 0.33 * (lats.max() - lats.min())
                    dlon = 0.33 * (lons.max() - lons.min())
                    self.grid_size["lats"] = np.linspace(
                        lats.min() - dlat, lats.max() + dlat, 20
                    )
                    self.grid_size["lons"] = np.linspace(
                        lons.min() - dlon, lons.max() + dlon, 25
                    )
                    tmp = gps2dist_azimuth(
                        self.grid_size["lats"].max(),
                        self.grid_size["lons"].max(),
                        self.grid_size["lats"].min(),
                        self.grid_size["lons"].min(),
                    )
                    max_depth = np.min([tmp[0] / 1000, 60.0])
                    self.grid_size["deps"] = np.linspace(0, max_depth, 10)

            self.calculate_traveltimes()

            if "kmps" in self.phase_types[0]:
                self._v0 = float(self.phase_types[0].split("kmps")[0])
            else:
                self._v0 = self.model.model.s_mod.v_mod.layers[0][4]
        else:
            self.load_traveltimes(tt_file)
            if "kmps" in self.phase_types[0]:
                self._v0 = float(self.phase_types[0].split("kmps")[0])
            else:
                self._v0 = self._model_layers[0][4]

        self.regrid = regrid
        self._normType = normType
        self._minimize = minimize_type
        self.plot = plot

        self.interact = interact
        if not self.plot:
            self.interact = False

        self._p = np.array(
            [4.9650, -18.9378, 26.7329, -16.5927, 3.8438]
        )  # weight polynomial coefficients
        self.Cmin = Cmin
        self._Cmax = Cmax
        self._dx_min = dx_min
        self.sta_min = sta_min
        self.bootstrap = bootstrap
        if bootstrap > 1:
            self.bootstrap = bootstrap + 1
        if self.bootstrap < 1 or self.bootstrap - np.round(self.bootstrap) != 0:
            print("Warning: 'bootstrap' value must be an integer > 0")
            print("Setting bootstrap = 1")
            self.bootstrap = 1

        self.bootstrap_prct = bootstrap_prct
        if lookup_type not in [
            "linear",
            "nearest",
            "zero",
            "slinear",
            "quadratic",
            "cubic",
        ]:
            print(
                "'lookup_type' needs to be: 'linear', 'nearest', 'zero', 'slinear', 'quadratic' or 'cubic'."
            )
            print("Setting lookup_type to 'cubic'.")
            lookup_type = "cubic"
        self.lookup_type = lookup_type

        self._dt = dt

        if dTmax_s:
            self.dTmax_s = dTmax_s
        else:
            seconds = [(tr.stats.npts - 1) / tr.stats.sampling_rate for tr in st]
            self.dTmax_s = np.min(
                [
                    0.5 * np.min(seconds),
                    xcorr_utils.station_distances(st).max() / self._v0 + self._dt,
                ]
            )

        dTmax = np.round(
            float(self.dTmax_s) * self.traces[0].stats.sampling_rate
        )  # can't make a total shift of more than dTmax samples
        # self._mlag         = int(2*dTmax) # change on 0ct-30-2017
        self._mlag = int(dTmax)

        if raw_traces:
            self.raw_traces = raw_traces
            if rd_freq:
                self.rd_freq = rd_freq

        if env_hp:
            self.env_hp = env_hp
            for tr in self.env_hp:
                tr.TT = self.traces.select(id=tr.id)[0].TT

        self._edge_control = edge_control
        self._num_processors = num_processors
        self._waveform_loc = waveform_loc
        self._gap_value = gap_value

    def __repr__(self):
        a = self.__dict__.keys()
        A = []
        for item in a:
            A.append(item + "=" + str(self.__getattribute__(item)))
        return "parameters(" + "\n".join(A) + ")"

    def calculate_traveltimes(self):
        """
        Calculate travel times based on the model and grid provided in the :class:`XCOR` object

        """

        if "default_vel_model" in self._model_file:
            try:
                self.model = TauPyModel(model=self._model_dir + "/default_vel_model")
            except:
                from obspy.taup.taup_create import build_taup_model

                build_taup_model(self._model_file, output_folder=self._model_dir)
                self._model_file = (
                    self._model_dir
                    + "/"
                    + self._model_file.split(".tvel")[0].split("/")[-1]
                )
                self.model = TauPyModel(model=self._model_file)
        else:
            if self._model_file.split(".")[-1] == "tvel":
                from obspy.taup.taup_create import build_taup_model

                build_taup_model(self._model_file, output_folder=self._model_dir)
                self._model_file = (
                    self._model_dir
                    + "/"
                    + self._model_file.split(".tvel")[0].split("/")[-1]
                )

            self.model = TauPyModel(model=self._model_file)

        self._grid = dict({"LON": [], "LAT": [], "DEP": []})
        if self.rotation:
            self._grid["LON"], self._grid["LAT"], self._grid["DEP"] = (
                xloc_utils.create_rotated_grid(self.rotation)
            )
        else:
            self._grid["LON"], self._grid["LAT"], self._grid["DEP"] = np.meshgrid(
                self.grid_size["lons"], self.grid_size["lats"], self.grid_size["deps"]
            )

        if self.output > 1:
            progress(0, len(self.grid_size["deps"]), status="complete")
        """
		if self.tt_calc=='full':
			for i,tr in enumerate(self.traces):
				TIMES=[]
				for j,lon in enumerate(self._grid['LON'].flatten()):
					deg=locations2degrees(tr.stats.coordinates.latitude,tr.stats.coordinates.longitude,self._grid['LAT'].flatten()[j],lon)
					PATHS=self.model.get_ray_paths(source_depth_in_km=self._grid['DEP'].flatten()[j],distance_in_degree=deg,receiver_depth_in_km=0,phase_list=self.phase_types)
					times=np.array([p.time for p in PATHS])
					TIMES.append(times.min())

				tr.TT=np.reshape(TIMES,self._grid['LON'].shape)
				if self.output:
					progress(i+1, len(self.traces), status='complete')
		else: 
		"""
        if self.rotation:
            LON = self._grid["LON"][:, :, 0]
            LAT = self._grid["LAT"][:, :, 0]
        else:
            LON, LAT = np.meshgrid(self.grid_size["lons"], self.grid_size["lats"])

        lat_corners = LAT[:: LAT.shape[0] - 1, :: LAT.shape[1] - 1]
        lon_corners = LON[:: LON.shape[0] - 1, :: LON.shape[1] - 1]
        dist = []
        for tr in self.traces:
            tr.TT = np.zeros(LAT.shape + (len(self.grid_size["deps"]),))
            for y, x in zip(lat_corners.flatten(), lon_corners.flatten()):
                dist.append(
                    locations2degrees(
                        tr.stats.coordinates.latitude,
                        tr.stats.coordinates.longitude,
                        y,
                        x,
                    )
                )

        max_deg = np.array(dist).max()
        DEGS = np.linspace(0, max_deg, 50)
        for i, z in enumerate(self.grid_size["deps"]):

            TIMES = []
            for x in DEGS:
                if x == 0 and z == 0:
                    TIMES.append(0)
                    continue
                try:
                    PATHS = self.model.get_ray_paths(
                        source_depth_in_km=z,
                        distance_in_degree=x,
                        receiver_depth_in_km=0,
                        phase_list=self.phase_types,
                    )
                    if not PATHS and len(self.phase_types) == 1:
                        PATHS = self.model.get_ray_paths(
                            source_depth_in_km=z,
                            distance_in_degree=x,
                            receiver_depth_in_km=0,
                            phase_list=[self.phase_types[0].lower()],
                        )
                except:
                    # an error can occur when a grid node is on a velocity model boundary. When this occurs,
                    # add a slight deviation to make sure get_ray_paths runs correctly
                    small_perturbation = 0.00001
                    PATHS = self.model.get_ray_paths(
                        source_depth_in_km=z + small_perturbation,
                        distance_in_degree=x,
                        receiver_depth_in_km=0,
                        phase_list=self.phase_types,
                    )
                    if not PATHS and len(self.phase_types) == 1:
                        PATHS = self.model.get_ray_paths(
                            source_depth_in_km=z,
                            distance_in_degree=x,
                            receiver_depth_in_km=0,
                            phase_list=[self.phase_types[0].lower()],
                        )
                if self.output > 2:
                    print(self.phase_types)
                    print(PATHS)
                times = np.array([p.time for p in PATHS])
                TIMES.append(times.min())

            for tr in self.traces:
                deg = np.array(
                    [
                        locations2degrees(
                            tr.stats.coordinates.latitude,
                            tr.stats.coordinates.longitude,
                            y,
                            x,
                        )
                        for y, x in zip(LAT.flatten(), LON.flatten())
                    ]
                )
                inter_times = np.interp(deg, DEGS, np.array(TIMES))
                TIM = np.reshape(inter_times, LAT.shape)
                tr.TT[:, :, i] = TIM

            if self.output > 1:
                progress(i + 1, len(self.grid_size["deps"]), status="complete")
        print("\n")

    def save_traveltimes(self, outfile):
        """
        Save an .npz file containing the model and grid. The variables should be created
        using the :meth:`calculate_traveltimes()` method, which is called internally  when initiating
        an :class:`XCOR` object with a model and grid.

        Parameters
        ----------
        file : str, required
                The .npz file you wish to save, ideally with full path provided.

        """

        model = self.model.model.s_mod.v_mod.layers
        phase_types = self.phase_types
        deps = self.grid_size["deps"]

        if self.rotation:
            Xs = self.grid_size["x"]
            Ys = self.grid_size["y"]
            lat0 = self.rotation["lat0"]
            lon0 = self.rotation["lon0"]
            az = self.rotation["az"]
        else:
            lats = self.grid_size["lats"]
            lons = self.grid_size["lons"]

        stas = {}
        for tr in self.traces:
            stas[tr.id.replace(".", "_")] = tr.TT

        if self.rotation:
            np.savez(
                outfile,
                y=Ys,
                x=Xs,
                deps=deps,
                az=az,
                lat0=lat0,
                lon0=lon0,
                model=model,
                phase_types=phase_types,
                stations=stas,
                **stas
            )
        else:
            np.savez(
                outfile,
                lats=lats,
                lons=lons,
                deps=deps,
                model=model,
                phase_types=phase_types,
                stations=stas,
                **stas
            )

    def load_traveltimes(self, file):
        """
        Load an .npz file containing a model and grid, and assign the appropriate times
        to each trace in the object. The .npz file can be created initially using the
        :meth:`save_traveltimes()` method after :class:`XCOR` has run :meth:`calculate_traveltimes()` with a model and grid.

        Parameters
        ----------
        file : str, required
                The path/filename to the .npz file you wish to load

        """
        npzfile = np.load(file)
        self._grid = dict({"LON": [], "LAT": [], "DEP": []})
        if self.rotation:
            self.grid_size = {
                "x": npzfile["x"],
                "y": npzfile["y"],
                "deps": npzfile["deps"],
            }
            self._grid["LON"], self._grid["LAT"], self._grid["DEP"] = (
                xloc_utils.create_rotated_grid(self.rotation)
            )
        else:
            self.grid_size = {
                "lons": npzfile["lons"],
                "lats": npzfile["lats"],
                "deps": npzfile["deps"],
            }
            self._grid["LON"], self._grid["LAT"], self._grid["DEP"] = np.meshgrid(
                self.grid_size["lons"], self.grid_size["lats"], self.grid_size["deps"]
            )
        self._model_layers = npzfile["model"]
        self.phase_types = npzfile["phase_types"].tolist()
        for tr in self.traces:
            tr.TT = npzfile[tr.id.replace(".", "_")]
            if np.shape(tr.TT) != np.shape(self._grid["LON"]):
                print("Error loading file. Dimension mismatch for station " + tr.id)

    def locate(
        self,
        window_list=[],
        window_length=None,
        step=None,
        offset=0,
        include_partial_windows=False,
        nearest_sample=True,
        dTmax_s=None,
        num_processors=None,
    ):
        """
        Main method to locate data provided. This method is actually a wrapper to call main envelope
        cross correlation location algorithm :meth:`XC_locate()`.

        If using this to locate a single window of data, simply call :meth:`locate()` with no input parameters.
        All of the relevant details (plot, interact, map_resolution) are properties set in the initial
        object creation.

        If using this to iterate over a list of user-provided windows, the following parameter must be set:


        Parameters
        ----------
        window_list : list, optional
                list containting 2x tuples of start and end times (Obspy UTCDatetime format) generated externally


        If using this to create and iterate over windows internally, the following parameters must be set:

        Parameters
        ----------
        window_length : float, optional
                Length of each sub-window in seconds.
                Default = `None`
        step : float, optional
                The step between the start times of two successive windows in seconds. Can be negative if an offset is given.
                Default = `None`


        These parameters are optional:

        Parameters
        ----------
        offset : float, optional
                The offset of the first window in seconds relative to the start time of the whole interval.
                Default = 0
        include_partial_windows : bool, optional
                Determines if sub-windows that are shorter then 99.9 % of the desired length are returned.
                Default = `False`
        nearest_sample : bool, optional
                If set to True, the closest sample is selected, if set to False, the inner (next sample for a
                start time border, previous sample for an end time border) sample containing the time is selected.
                Default = `True`
        dTmax_s : float, optional
                Maximum cross-correlation shift in seconds. Defaults to the smaller of

                a) 1/2 of the obspy stream sub-window length

                b) the maximum predicted inter-station differential time + 'dt', set during initial object creation
        num_processors : int, optional
                Number of processors to use in parallel
                Default = `None`

        """
        if num_processors:
            self._num_processors = num_processors
        if not window_length and not window_list:
            if dTmax_s:
                self.dTmax_s = dTmax_s
                dTmax = np.round(
                    float(self.dTmax_s) * self.traces[0].stats.sampling_rate
                )  # can't make a total shift of more than dTmax samples
                # self._mlag         = int(2*dTmax) # change on 0ct-30-2017
                self._mlag = int(dTmax)
            loc = XC_locate(
                (self.traces[0].stats.starttime, self.traces[0].stats.endtime), self
            )
        else:
            self._windows = True
            if window_length:
                if not step:
                    print("Set step size")
                    return

                from obspy.core.util.misc import get_window_times

                windows = get_window_times(
                    starttime=self.traces[0].stats.starttime,
                    endtime=self.traces[0].stats.endtime,
                    window_length=window_length,
                    step=step,
                    offset=offset,
                    include_partial_windows=include_partial_windows,
                )
            elif window_list:
                windows = window_list

            if dTmax_s:
                self.dTmax_s = dTmax_s
            else:
                win_len = windows[0][1] - windows[0][0]
                self.dTmax_s = np.min(
                    [
                        0.5 * win_len,
                        xcorr_utils.station_distances(self.traces).max() / self._v0
                        + self._dt,
                    ]
                )
            dTmax = np.round(
                float(self.dTmax_s) * self.traces[0].stats.sampling_rate
            )  # can't make a total shift of more than dTmax samples
            # self._mlag         = int(2*dTmax) # change on 0ct-30-2017
            self._mlag = int(dTmax)

            bp_traces = self.traces.copy()
            if "env_hp" in self.__dict__:
                env_hp = True
                hp_traces = self.env_hp.copy()
                self.env_hp = []
            else:
                env_hp = False

            if "raw_traces" in self.__dict__:
                rd_flag = True
                raw_traces = self.raw_traces.copy()
                self.__delattr__("raw_traces")
            else:
                rd_flag = False

            if self._num_processors > 1:
                from multiprocessing import Pool

                pool = Pool(processes=self._num_processors)
                results = [
                    pool.apply_async(XC_locate, args=(win, self)) for win in windows
                ]

                pool.close()
                pool.join()
                loc = event_list([p.get() for p in results])

                self.traces = []
                if env_hp:
                    new_windows = []
                    for i, l in enumerate(loc.events):
                        l.highpass_loc = False
                        if not np.isnan(l.latitude):
                            new_windows.append(windows[i])
                    if self.output > 0:
                        print(
                            "Calculating highpass locations for {:.0f} locations".format(
                                len(new_windows)
                            )
                        )
                    tmp_bootstrap = copy(self.bootstrap)
                    tmp_regrid = copy(self.regrid)
                    self.traces = hp_traces
                    self.bootstrap = 10
                    self.regrid = False

                    pool = Pool(processes=self._num_processors)
                    results = [
                        pool.apply_async(XC_locate, args=(win, self))
                        for win in new_windows
                    ]
                    pool.close()
                    pool.join()
                    # loc_hp=event_list([p.get() for p in results]).remove(max_scatter=10)
                    loc_hp = event_list([p.get() for p in results])
                    print(len(loc_hp.events))
                    loc_hp.remove(
                        max_scatter=5, rm_nan_loc=True, rm_nan_err=True, inplace=True
                    )
                    print(len(loc_hp.events))
                    times = np.array(loc.get_times()[0])
                    for t in loc_hp.get_times()[0]:
                        ind = np.where(times == t)[0][0]
                        loc.events[ind].highpass_loc = True
                    self.traces = []

                if rd_flag:
                    pool = Pool(processes=self._num_processors)
                    new_windows = []
                    for i, l in enumerate(loc.events):
                        if not np.isnan(l.latitude):
                            new_windows.append((windows[i], l))
                    if self.output > 0:
                        print(
                            "Calculating reduced displacement for {:.0f} locations".format(
                                len(new_windows)
                            )
                        )
                    results = [
                        pool.apply_async(
                            xloc_utils.reduced_displacement,
                            args=(win[1], self, raw_traces.slice(win[0][0], win[0][1])),
                        )
                        for win in new_windows
                    ]
                    pool.close()
                    pool.join()
                    RD = [p.get() for p in results]
                    if not env_hp:
                        times = np.array(loc.get_times()[0])
                    for i, rd in enumerate(RD):
                        ind = np.where(times == new_windows[i][0][0])[0][0]
                        loc.events[ind].reduced_displacement = rd
                    self.raw_traces = raw_traces

            else:
                loc = event_list([XC_locate(win, self) for win in windows])
                if env_hp:
                    if self.output > 0:
                        print("Calculating highpass locations...")
                    tmp_bootstrap = copy(self.bootstrap)
                    tmp_regrid = copy(self.regrid)
                    self.traces = hp_traces
                    self.bootstrap = 10
                    self.regrid = False
                    loc_hp = []
                    for i, l in enumerate(loc.events):
                        l.highpass_loc = False
                        if not np.isnan(l.latitude):
                            loc_hp.append(XC_locate(windows[i], self))
                    loc_hp = event_list(loc_hp)
                    loc_hp.remove(
                        max_scatter=5, rm_nan_loc=True, rm_nan_err=True, inplace=True
                    )
                    times = np.array(loc.get_times()[0])
                    for t in loc_hp.get_times()[0]:
                        ind = np.where(times == t)[0][0]
                        loc.events[ind].highpass_loc = True
                    self.traces = []
                if rd_flag:
                    new_windows = []
                    for i, l in enumerate(loc.events):
                        if not np.isnan(l.latitude):
                            new_windows.append((windows[i], l))
                    if self.output > 0:
                        print(
                            "Calculating reduced displacement for {:.0f} locations".format(
                                len(new_windows)
                            )
                        )
                    RD = [
                        xloc_utils.reduced_displacement(
                            win[1], self, raw_traces.slice(win[0][0], win[0][1])
                        )
                        for win in new_windows
                    ]
                    if not env_hp:
                        times = np.array(loc.get_times()[0])
                    for i, rd in enumerate(RD):
                        ind = np.where(times == new_windows[i][0][0])[0][0]
                        loc.events[ind].reduced_displacement = rd
                    self.raw_traces = raw_traces

            self.traces = bp_traces
            if env_hp:
                self.env_hp = hp_traces
                self.bootstrap = tmp_bootstrap
                self.regrid = tmp_regrid

        return loc

    def plot_grid(self):
        from enveloc import plotting_utils

        plotting_utils.grid_plot(self)



from obspy import UTCDateTime, Stream
from obspy.signal.filter import envelope
from obspy.core.util import AttribDict
from obspy.clients.fdsn import Client
import numpy as np
from enveloc.core import XCOR

client = Client("IRIS")

dt = 10


def add_latlon(st, client=None):

    if not client:
        client = Client("IRIS")
    for tr in st:
        print(tr)
        inventory = client.get_stations(
            network=tr.stats.network,
            station=tr.stats.station,
            location=tr.stats.location,
            channel=tr.stats.channel,
            starttime=tr.stats.starttime,
            endtime=tr.stats.endtime,
        )
        tr.stats.coordinates = AttribDict(
            {
                "latitude": inventory[0][0].latitude,
                "longitude": inventory[0][0].longitude,
                "elevation": inventory[0][0].elevation,
            }
        )
    return st


def get_data(sta_list, t1, t2, fill_value=0, client=None):

    st = Stream()

    if not client:
        client = Client("IRIS")
    for channel in sta_list:
        net, sta, loc, chan = channel.split(".")
        try:
            tr = client.get_waveforms(
                net,
                sta,
                loc,
                chan,
                UTCDateTime(t1) - dt,
                UTCDateTime(t2) + dt,
                attach_response=True,
            )
            if len(tr) > 1:
                if fill_value == 0 or fill_value == None:
                    tr.detrend("demean")
                    tr.taper(max_percentage=None, max_length=1)
                for sub_trace in tr:
                    # deal with rare error when sub-traces have different dtypes
                    if sub_trace.data.dtype.name != "int32":
                        sub_trace.data = sub_trace.data.astype("int32")
                    if sub_trace.data.dtype != np.dtype("int32"):
                        sub_trace.data = sub_trace.data.astype("int32")
                    # deal with rare error when sub-traces have different sample rates
                    if sub_trace.stats.sampling_rate != round(
                        sub_trace.stats.sampling_rate
                    ):
                        sub_trace.stats.sampling_rate = round(
                            sub_trace.stats.sampling_rate
                        )
                print("Merging gappy data...")
                tr.merge(fill_value=fill_value)
        except:
            print("Cannot get data for {}".format(channel))
            continue
        st += tr

    return st


def make_env(st, lowpass=0.2):

    st.detrend("demean")
    for tr in st:
        tr.resample(25.0)
        if tr.stats.npts % 2 == 1:
            tr.trim(
                starttime=tr.stats.starttime,
                endtime=tr.stats.endtime + 1 / tr.stats.sampling_rate,
                pad=True,
                fill_value=0,
            )
        tr.data = envelope(tr.data)
        tr.resample(5.0)

    st.filter("lowpass", freq=lowpass)

    return st


def get_IRIS_data(sta_list, t1, t2, f1=1.0, f2=8.0, lowpass=0.2, client=None):

    if not client:
        client = Client("IRIS")
    print("Downloading data...\n")
    st = get_data(sta_list, t1, t2, client=client)
    print("Adding lat/lon info...\n")
    st = add_latlon(st, client=client)
    print("Detrending data...\n")
    st.detrend("demean")
    print("Tapering data...\n")
    st.taper(max_percentage=None, max_length=5)
    print("Filtering data...\n")
    st.filter("bandpass", freqmin=f1, freqmax=f2, corners=3, zerophase=True)
    print("Making envelopes...\n")
    env = make_env(st.copy(), lowpass=lowpass)
    print("Final trim...\n")
    env.trim(env[0].stats.starttime + dt, env[0].stats.endtime - dt)

    return env


def interactive_example():

    t1 = "2018-04-28 13:07"
    t2 = "2018-04-28 13:09"

    FREQMIN = 1.0
    FREQMAX = 8.0
    LOWPASS = 0.2

    sta_list = [
        "HV.BYL..HHZ",
        "HV.DEVL..HHZ",
        "HV.HAT..HHZ",
        "HV.KKO..HHZ",
        "HV.NPT..HHZ",
        "HV.OBL..HHZ",
        "HV.PAUD..HHZ",
        "HV.PUHI..HHZ",
        "HV.RIMD..HHZ",
        "HV.SBL..HHZ",
        "HV.SDH..HHZ",
        "HV.UWB..HHZ",
        "HV.UWE..HHZ",
        "HV.WRM..HHZ",
    ]

    env = get_IRIS_data(
        sta_list, t1, t2, f1=FREQMIN, f2=FREQMAX, lowpass=LOWPASS, client=client
    )

    print("Creating XCOR object....\n")
    XC = XCOR(env)
    print("Locating from envelopes:\n")
    loc = XC.locate()

    return loc, XC


def test():

    t1 = "2018-04-28 13:07"
    t2 = "2018-04-28 13:09"

    FREQMIN = 1.0
    FREQMAX = 8.0
    LOWPASS = 0.2

    sta_list = [
        "HV.BYL..HHZ",
        "HV.DEVL..HHZ",
        "HV.HAT..HHZ",
        "HV.KKO..HHZ",
        "HV.NPT..HHZ",
        "HV.OBL..HHZ",
        "HV.PAUD..HHZ",
        "HV.PUHI..HHZ",
        "HV.RIMD..HHZ",
        "HV.SBL..HHZ",
        "HV.SDH..HHZ",
        "HV.UWB..HHZ",
        "HV.UWE..HHZ",
        "HV.WRM..HHZ",
    ]

    print("Downloading some data for test...")
    env = get_IRIS_data(
        sta_list, t1, t2, f1=FREQMIN, f2=FREQMAX, lowpass=LOWPASS, client=client
    )

    print("Creating XCOR object....\n")
    XC = XCOR(env, plot=True, interact=False)
    print("Locating from envelopes:\n")
    loc = XC.locate()


def custom_grid_example():

    t1 = "2018-04-28 13:07"
    t2 = "2018-04-28 13:09"

    FREQMIN = 1.0
    FREQMAX = 8.0
    LOWPASS = 0.2

    sta_list = [
        "HV.BYL..HHZ",
        "HV.DEVL..HHZ",
        "HV.HAT..HHZ",
        "HV.KKO..HHZ",
        "HV.NPT..HHZ",
        "HV.OBL..HHZ",
        "HV.PAUD..HHZ",
        "HV.PUHI..HHZ",
        "HV.RIMD..HHZ",
        "HV.SBL..HHZ",
        "HV.SDH..HHZ",
        "HV.UWB..HHZ",
        "HV.UWE..HHZ",
        "HV.WRM..HHZ",
    ]

    env = get_IRIS_data(
        sta_list, t1, t2, f1=FREQMIN, f2=FREQMAX, lowpass=LOWPASS, client=client
    )

    print("Creating XCOR object....\n")

    mygrid = {
        "deps": np.arange(0, 14, 0.5),
        "lons": np.arange(-155.35, -155.2, 0.002),
        "lats": np.arange(19.35, 19.45, 0.002),
    }

    XC = XCOR(env, grid_size=mygrid, interact=False)
    print("Locating from envelopes:\n")
    loc = XC.locate()

    return loc, XC


def rotate_grid_example():

    t1 = "2018-04-28 13:07"
    t2 = "2018-04-28 13:09"

    FREQMIN = 1.0
    FREQMAX = 8.0
    LOWPASS = 0.2

    sta_list = [
        "HV.BYL..HHZ",
        "HV.DEVL..HHZ",
        "HV.HAT..HHZ",
        "HV.KKO..HHZ",
        "HV.NPT..HHZ",
        "HV.OBL..HHZ",
        "HV.PAUD..HHZ",
        "HV.PUHI..HHZ",
        "HV.RIMD..HHZ",
        "HV.SBL..HHZ",
        "HV.SDH..HHZ",
        "HV.UWB..HHZ",
        "HV.UWE..HHZ",
        "HV.WRM..HHZ",
    ]

    env = get_IRIS_data(
        sta_list, t1, t2, f1=FREQMIN, f2=FREQMAX, lowpass=LOWPASS, client=client
    )

    print("Creating XCOR object....\n")

    rotation = {
        "x": np.arange(-5, 5, 0.3),
        "y": np.arange(-3.5, 3.5, 0.3),
        "z": np.arange(0, 25, 2),
        "lat0": 19.403,
        "lon0": -155.281,
        "az": 30,
    }

    XC = XCOR(env, rotation=rotation, interact=False)
    print("Locating from envelopes:\n")
    loc = XC.locate()

    return loc, XC


def auto_example():

    t1 = "2017-05-01 13:00"
    t2 = "2017-05-01 15:00"

    FREQMIN = 0.5
    FREQMAX = 10
    LOWPASS = 0.2

    sta_list = [
        "HV.BYL..HHZ",
        "HV.DEVL..HHZ",
        "HV.HAT..HHZ",
        "HV.KKO..HHZ",
        "HV.NPT..HHZ",
        "HV.OBL..HHZ",
        "HV.PAUD..HHZ",
        "HV.PUHI..HHZ",
        "HV.RIMD..HHZ",
        "HV.SBL..HHZ",
        "HV.SDH..HHZ",
        "HV.UWB..HHZ",
        "HV.UWE..HHZ",
        "HV.WRM..HHZ",
    ]

    env = get_IRIS_data(
        sta_list, t1, t2, f1=FREQMIN, f2=FREQMAX, lowpass=LOWPASS, client=client
    )

    print("Creating XCOR object....\n")
    XC = XCOR(env, bootstrap=20, plot=False, output=2)
    print("Locating from envelopes:\n")
    locs = XC.locate(window_length=240, step=120)

    return locs, env


def cascadia_example():

    t1 = "2020-05-24 04:52:30"
    t2 = "2020-05-24 05:07:30"

    FREQMIN = 1.5
    FREQMAX = 6.0
    LOWPASS = 0.1

    sta_list = [
        "UW.MCW.01.EHZ",
        "PB.B011.--.EHZ",
        "CN.SYMB.--.HHZ",
        "CN.PTRF.--.HHZ",
        "UW.RPW.01.EHZ",
        "CN.VGZ.--.HHZ",
        "PB.B004.--.EHZ",
        "UW.JCW.--.EHZ",
        "UW.SQM.--.EHZ",
        "PB.B003.--.EHZ",
        "PB.B006.--.EHZ",
        "PB.B001.--.EHZ",
        "UW.OBC.--.EHZ",
        "PB.B013.--.EHZ",
        "UW.DOSE.--.HHZ",
        "UW.HDW.--.EHZ",
        "UW.GNW.--.HHZ",
        "UW.GMW.--.EHZ",
        "PB.B014.--.EHZ",
        "UW.SMW.--.EHZ",
        "UW.STOR.--.HHZ",
        "UW.TKEY.--.HHZ",
    ]

    env = get_IRIS_data(
        sta_list, t1, t2, f1=FREQMIN, f2=FREQMAX, lowpass=LOWPASS, client=client
    )

    mygrid = {
        "lons": np.arange(-125, -121 + 0.05, 0.075),
        "lats": np.arange(46.5, 49.0 + 0.05, 0.075),
        "deps": np.arange(20, 60 + 0.1, 10),
    }

    print("Creating XCOR object....\n")
    XC = XCOR(env, grid_size=mygrid, regrid=True, bootstrap=30, output=2)
    print("Locating from envelopes:\n")
    loc = XC.locate()

    return loc, XC



from collections import Counter
import matplotlib.pyplot as plt
from matplotlib import gridspec
import matplotlib.dates as mdates
from matplotlib.widgets import Button
import numpy as np

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.ticker as mticker
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter


trace_ax = []
h_trace = []
h_5 = []
h_maptext = []
h_mapdot = []
fig = []
krm = []
sort_key = []

relocate = False
restart = False
plot_opt = dict({"krm": krm, "relocate": relocate, "restart": restart})

st_dot_color0 = "y"
st_dot_color1 = "darkorange"
st_text_color0 = "darkblue"
st_text_color1 = "darkred"
trace_color0 = "k"
trace_color1 = "r"
xcorr_color0 = "royalblue"
xcorr_color1 = "r"

interval = 0.05


def on_click(event):
    global krm
    if event.inaxes == trace_ax:
        ind = int(np.round(event.ydata))
        if ind > len(h_trace) - 1:
            return
        if ind < 0:
            return
        if ind in krm:
            krm.remove(ind)
            h_trace[ind][0].set_color(trace_color0)
            h_trace[ind][0].set_linewidth(0.2)
            h_mapdot[ind][0].set_color(st_dot_color0)
            h_mapdot[ind][0].set_zorder(100)
            h_maptext[ind].set_color(st_text_color0)
            h_maptext[ind].set_zorder(100)
            # h_maptext[ind].set_weight('normal')
            for t in h_5[ind]:
                t[0].set_color(xcorr_color0)
                t[0].set_linewidth(0.5)
                t[0].set_zorder(100)

        else:
            krm.append(ind)
            h_trace[ind][0].set_color(trace_color1)
            h_trace[ind][0].set_linewidth(0.5)
            h_mapdot[ind][0].set_color(st_dot_color1)
            h_mapdot[ind][0].set_zorder(1000)
            h_maptext[ind].set_color(st_text_color1)
            h_maptext[ind].set_zorder(1000)
            # h_maptext[ind].set_weight('bold')
            for t in h_5[ind]:
                t[0].set_color(xcorr_color1)
                t[0].set_linewidth(1.0)
                t[0].set_zorder(1000)

        fig.canvas.draw()


class Index(object):

    def done(self, event):
        global krm
        global relocate
        global restart
        krm = []
        restart = False
        relocate = False
        plt.pause(interval)
        plt.close()

    def reloc(self, event):
        global relocate
        global restart
        restart = False
        relocate = True
        plt.pause(interval)
        plt.close()

    def restart(self, event):
        global relocate
        global restart
        restart = True
        relocate = False
        plt.pause(interval)
        plt.close()


def XC_plot(CC, XC, CC1, misfit, loc):
    del krm[:]
    plot_opt = dict({"krm": [], "relocate": False, "restart": False})
    import warnings

    warnings.filterwarnings("ignore")

    global fig
    global trace_ax
    global h_trace
    global h_5
    global h_maptext
    global h_mapdot
    global sort_key

    trace_ax = []
    h_trace = []
    h_5 = []
    h_maptext = []
    h_mapdot = []
    fig = []

    ######## calculate travel times ########
    ########################################
    st_sort = CC["st"].copy()
    Nseis0 = len(CC["st"])
    dTs = np.zeros((Nseis0,))
    ind_min = misfit.argmin()
    for ksta in range(Nseis0):
        dTs[ksta] = CC["st"][ksta].TT.flatten(order="F")[ind_min]
        st_sort[ksta].dTs = dTs[ksta]
        st_sort[ksta].staW = CC["staW"][ksta]
    dT = dTs * CC["st"][0].stats.sampling_rate
    deltaT = dT[CC["j1"]] - dT[CC["i1"]] + XC._mlag
    sort_key = np.argsort(dTs)[::-1]
    st_sort.traces.sort(key=lambda x: x.dTs, reverse=True)
    ########################################
    ########################################

    gs = gridspec.GridSpec(60, 20)
    fig = plt.figure(num="XC loc", figsize=(6, 7.2))

    ############### plot map ###############
    ########################################
    map_ax = fig.add_subplot(gs[:25, :13], projection=ccrs.PlateCarree())
    if XC.rotation:
        latlims = [XC._grid["LAT"].flatten().min(), XC._grid["LAT"].flatten().max()]
        lonlims = [XC._grid["LON"].flatten().min(), XC._grid["LON"].flatten().max()]
    else:
        latlims = [XC.grid_size["lats"][0], XC.grid_size["lats"][-1]]
        lonlims = [XC.grid_size["lons"][0], XC.grid_size["lons"][-1]]
    map_ax.set_extent(lonlims + latlims, crs=ccrs.PlateCarree())
    map_ax.add_feature(cfeature.LAND, color="silver")
    map_ax.add_feature(cfeature.OCEAN, color="lightblue")
    map_ax.add_feature(cfeature.COASTLINE)
    map_ax.add_feature(cfeature.BORDERS, linestyle=":")
    map_ax.add_feature(cfeature.LAKES, color="lightblue", alpha=0.5)
    map_ax.add_feature(cfeature.RIVERS, color="lightblue")

    # add lat/lon labels
    gl = map_ax.gridlines(
        crs=ccrs.PlateCarree(),
        draw_labels=True,
        linewidth=0.5,
        color="white",
        alpha=0.8,
        linestyle=":",
    )
    gl.xformatter = LongitudeFormatter(zero_direction_label=True)
    gl.yformatter = LatitudeFormatter()
    meridians = np.linspace(lonlims[0], lonlims[-1], 4)
    parallels = np.linspace(latlims[0], latlims[-1], 4)
    gl.xlocator = mticker.FixedLocator(meridians)
    gl.ylocator = mticker.FixedLocator(parallels)
    gl.bottom_labels = False
    gl.right_labels = False
    gl.xlabel_style = {"size": 6}
    gl.ylabel_style = {"size": 6}

    # add stations to map
    stalats = np.array([tr.stats.coordinates.latitude for tr in st_sort])
    stalons = np.array([tr.stats.coordinates.longitude for tr in st_sort])
    stanames = [tr.stats.station for tr in st_sort]
    for i, name in enumerate(stanames):
        h_mapdot.append(
            map_ax.plot(
                stalons[i],
                stalats[i],
                transform=ccrs.PlateCarree(),
                marker="^",
                color=st_dot_color0,
                markeredgecolor="k",
                markersize=8,
                markeredgewidth=0.5,
            )
        )
        h_maptext.append(
            map_ax.text(stalons[i], stalats[i], name, fontsize=9, color=st_text_color0)
        )

    # add location(s) to the map
    rand1 = (np.random.random_sample((len(loc.bstrap_lon),)) - 0.5) / 50.0
    rand2 = (np.random.random_sample((len(loc.bstrap_lon),)) - 0.5) / 50.0
    map_ax.plot(
        loc.bstrap_lon + rand1,
        loc.bstrap_lat + rand2,
        ".",
        color="lightgreen",
        markeredgecolor="k",
        markersize=2,
        markeredgewidth=0.5,
    )
    map_ax.plot(
        loc.longitude,
        loc.latitude,
        "*",
        markerfacecolor="firebrick",
        markeredgecolor="k",
        markersize=10,
        markeredgewidth=0.5,
    )

    if np.isnan(loc.horizontal_scatter):
        plt.title(
            "Lat: {:.3f}, Lon: {:.3f}\nDepth {:.1f} km".format(
                loc.latitude, loc.longitude, loc.depth
            ),
            fontsize=8,
        )
    else:
        plt.title(
            "Lat: {:.3f}, Lon: {:.3f} (scatter: {:.1f} km)\nDepth {:.1f} km (scatter {:.1f} km)".format(
                loc.latitude,
                loc.longitude,
                loc.horizontal_scatter,
                loc.depth,
                loc.vertical_scatter,
            ),
            fontsize=8,
        )
    ########################################
    ########################################

    ############ plot envelopes ############
    ########################################
    trace_ax = plt.subplot(gs[30:, :9])
    for i, tr in enumerate(st_sort):
        T = np.linspace(
            mdates.date2num(tr.stats.starttime.datetime),
            mdates.date2num(tr.stats.endtime.datetime),
            len(tr.data),
        )
        h_trace.append(
            plt.plot(
                mdates.num2date(T),
                i + tr.data / tr.data.max(),
                color=trace_color0,
                linewidth=0.2,
            )
        )
        plt.text(
            mdates.date2num(tr.stats.endtime.datetime),
            i,
            " {:.1f}s, {:.1f}".format(tr.dTs, tr.staW / len(CC["st"])),
            fontsize=6,
        )
    plt.text(
        mdates.date2num(tr.stats.endtime.datetime), i + 1, " time, weight", fontsize=5
    )
    trace_ax.set_yticks(np.arange(len(st_sort)))
    trace_ax.set_yticklabels(
        [tr.stats.station + "." + tr.stats.channel for tr in st_sort]
    )

    plt.xticks(
        np.linspace(
            mdates.date2num(tr.stats.starttime.datetime),
            mdates.date2num(tr.stats.endtime.datetime),
            4,
        )
    )
    plt.tick_params(axis="both", which="major", labelsize=6)
    plt.xlim(T[0], T[-1])
    plt.ylim(-2, len(CC["st"]) + 0.25)
    xfmt = mdates.DateFormatter("%H:%M:%S")
    trace_ax.xaxis.set_major_formatter(xfmt)
    plt.grid(linestyle="--", linewidth=0.2, zorder=0, axis="x")
    plt.title(
        "Envelopes starting at\n{}".format(
            CC["st"][0].stats.starttime.strftime("%Y.%m.%d %H:%M:%S")
        ),
        fontsize=8,
    )
    ########################################
    ########################################

    ########## plot correlations ###########
    ########################################
    cc_ax = plt.subplot(gs[30:, 11:])
    h_5 = list()
    for i in range(len(CC["st"])):
        k0 = sort_key[i]
        tmp_list = list()
        ii = np.where(CC["j1"] == k0)[0]
        for sub_ind in ii:
            if CC["W2"][sub_ind] > 0:
                DT0 = CC["tC"][np.round(deltaT[sub_ind]).astype("int")]
                tmp_list.append(
                    plt.plot(
                        CC["tC"] - DT0,
                        np.abs(DT0) + (1 - CC1[:, sub_ind]) * 10,
                        "-",
                        color=xcorr_color0,
                        linewidth=0.5,
                        alpha=0.3,
                    )
                )
        ii = np.where(CC["i1"] == k0)[0]
        for sub_ind in ii:
            if CC["W2"][sub_ind] > 0:
                DT0 = CC["tC"][np.round(deltaT[sub_ind]).astype("int")]
                tmp_list.append(
                    plt.plot(
                        CC["tC"] - DT0,
                        np.abs(DT0) + (1 - CC1[:, sub_ind]) * 10,
                        "-",
                        color=xcorr_color0,
                        linewidth=0.5,
                        alpha=0.3,
                    )
                )
        h_5.append(tmp_list)

    plt.grid(linestyle="--", linewidth=0.2, zorder=0)
    plt.tick_params(axis="both", which="major", labelsize=6)
    plt.yticks([], [])
    plt.title("Time shifted cross-correlations (s)", fontsize=8)
    ########################################
    ########################################

    msft = np.reshape(misfit, np.shape(XC._grid["LON"]), order="F")
    ii = np.unravel_index(msft.argmin(), np.shape(XC._grid["LON"]), order="C")

    ########### plot misfit map ############
    ########################################
    misfit_ax1 = plt.subplot(gs[:11, 14:], projection=ccrs.PlateCarree())
    misfit_ax1.set_extent(lonlims + latlims, crs=ccrs.PlateCarree())
    misfit_ax1.add_feature(cfeature.COASTLINE)
    misfit_ax1.contourf(
        XC._grid["LON"][:, :, ii[2]],
        XC._grid["LAT"][:, :, ii[2]],
        msft[:, :, ii[2]],
        40,
        transform=ccrs.PlateCarree(),
        cmap="viridis_r",
    )
    msft_min = msft[:, :, ii[2]].min()
    msft_max = msft[:, :, ii[2]].max()
    misfit_ax1.plot(
        loc.longitude,
        loc.latitude,
        marker="*",
        color="firebrick",
        markeredgewidth=0.1,
        markeredgecolor="firebrick",
        markersize=6,
    )
    misfit_ax1.plot(
        stalons,
        stalats,
        marker="^",
        color="gray",
        markeredgewidth=0.1,
        markeredgecolor="gray",
        markersize=4,
        linewidth=0,
    )
    gl2 = misfit_ax1.gridlines(
        crs=ccrs.PlateCarree(),
        draw_labels=True,
        linewidth=0.2,
        color="white",
        alpha=0.8,
        linestyle=":",
    )
    gl2.xlocator = mticker.FixedLocator(meridians)
    gl2.ylocator = mticker.FixedLocator(parallels)
    gl2.yformatter = LatitudeFormatter()
    gl2.bottom_labels = False
    gl2.xlabels_top = False
    gl2.ylabels_left = False
    gl2.ylabel_style = {"size": 6}
    plt.title("Misfit Function", fontsize=8)
    ########################################
    ########################################

    ########## plot misfit slice ###########
    ########################################
    misfit_ax2 = plt.subplot(gs[12:24, 14:])
    map_aspect = (meridians[-1] - meridians[0]) / (
        (parallels[-1] - parallels[0]) * 111.1
    )
    if map_aspect > 1:
        misfit_ax2.set_aspect(
            (meridians[-1] - meridians[0]) / ((parallels[-1] - parallels[0]) * 111.1)
        )
    plt.contourf(
        XC._grid["LON"][ii[0], :, :],
        -XC._grid["DEP"][ii[0], :, :],
        msft[ii[0], :, :],
        40,
        cmap="viridis_r",
        vmin=msft_min,
        vmax=msft_max,
    )
    plt.plot(
        loc.longitude,
        -loc.depth,
        "*",
        markerfacecolor="firebrick",
        markeredgecolor="firebrick",
        markeredgewidth=0.1,
    )
    misfit_ax2.yaxis.tick_right()
    plt.ylabel("Depth", fontsize=8)
    misfit_ax2.yaxis.set_label_position("right")
    plt.gca().set_xticks(meridians[1:3])
    plt.xlim(lonlims)
    misfit_ax2.xaxis.set_major_formatter(LongitudeFormatter(zero_direction_label=True))
    plt.xlabel("Longitude", fontsize=8)
    plt.tick_params(axis="y", which="major", labelsize=6)
    plt.tick_params(axis="x", which="major", labelsize=5)
    plt.grid(linestyle=":", linewidth=0.2, color="w")
    ########################################
    ########################################

    # fig.subplots_adjust(left=0.1,bottom=0.04,right=0.91,top=0.95,wspace=0.0,hspace=0.0)
    # a1=misfit_ax1.get_position()
    # a2=misfit_ax2.get_position()
    # pos = [a1.bounds[0],a2.bounds[1],a1.width,a1.height]
    # misfit_ax2.set_position(pos)

    if XC.interact:
        callback = Index()
        ax_done = plt.axes([0.13, 0.04, 0.06, 0.02])
        b_done = Button(
            ax_done, label="Done", color="yellow", hovercolor="lightskyblue"
        )
        b_done.label.set_fontsize(6)
        b_done.on_clicked(callback.done)

        ax_reloc = plt.axes([0.24, 0.04, 0.08, 0.02])
        b_reloc = Button(
            ax_reloc, label="Relocate", color="yellow", hovercolor="lightskyblue"
        )
        b_reloc.label.set_fontsize(6)
        b_reloc.on_clicked(callback.reloc)

        ax_restart = plt.axes([0.365, 0.04, 0.08, 0.02])
        b_restart = Button(
            ax_restart, label="Restart", color="yellow", hovercolor="lightskyblue"
        )
        b_restart.label.set_fontsize(6)
        b_restart.on_clicked(callback.restart)

        cid = fig.canvas.mpl_connect("button_press_event", on_click)
        plt.show()
        plt.pause(0.5)
        fig.canvas.mpl_disconnect(cid)
        for i, ind in enumerate(krm):
            krm[i] = sort_key[ind]
    else:
        plt.show()

    plot_opt["krm"] = np.array(krm)
    plot_opt["relocate"] = relocate
    plot_opt["restart"] = restart

    return plot_opt


def grid_plot(XC):
    from obspy.geodetics.base import gps2dist_azimuth

    ############### plot map ###############
    ########################################

    fig = plt.figure(figsize=(6, 7.2))

    map_ax = fig.add_subplot(111, projection=ccrs.PlateCarree())
    if XC.rotation:
        latlims = [XC._grid["LAT"].flatten().min(), XC._grid["LAT"].flatten().max()]
        lonlims = [XC._grid["LON"].flatten().min(), XC._grid["LON"].flatten().max()]
    else:
        latlims = [XC.grid_size["lats"][0], XC.grid_size["lats"][-1]]
        lonlims = [XC.grid_size["lons"][0], XC.grid_size["lons"][-1]]
    map_ax.set_extent(lonlims + latlims, crs=ccrs.PlateCarree())
    map_ax.add_feature(cfeature.LAND, color="silver")
    map_ax.add_feature(cfeature.OCEAN, color="lightblue")
    map_ax.add_feature(cfeature.COASTLINE)
    map_ax.add_feature(cfeature.BORDERS, linestyle=":")
    map_ax.add_feature(cfeature.LAKES, color="lightblue", alpha=0.5)
    map_ax.add_feature(cfeature.RIVERS, color="lightblue")

    # add lat/lon labels
    gl = map_ax.gridlines(
        crs=ccrs.PlateCarree(),
        draw_labels=True,
        linewidth=0.5,
        color="white",
        alpha=0.8,
        linestyle=":",
    )
    gl.xformatter = LongitudeFormatter(zero_direction_label=True)
    gl.yformatter = LatitudeFormatter()
    meridians = np.linspace(lonlims[0], lonlims[-1], 4)
    parallels = np.linspace(latlims[0], latlims[-1], 4)
    gl.xlocator = mticker.FixedLocator(meridians)
    gl.ylocator = mticker.FixedLocator(parallels)
    gl.bottom_labels = False
    gl.right_labels = False
    gl.xlabel_style = {"size": 6}
    gl.ylabel_style = {"size": 6}

    # add grid nodes
    map_ax.plot(
        XC._grid["LON"][:, :, 0].flatten(),
        XC._grid["LAT"][:, :, 0].flatten(),
        transform=ccrs.PlateCarree(),
        marker="+",
        color="g",
        markersize=3,
        linewidth=0,
    )

    # add stations to map
    stalats = np.array([tr.stats.coordinates.latitude for tr in XC.traces])
    stalons = np.array([tr.stats.coordinates.longitude for tr in XC.traces])
    stanames = [tr.stats.station for tr in XC.traces]
    for i, name in enumerate(stanames):
        map_ax.plot(
            stalons[i],
            stalats[i],
            transform=ccrs.PlateCarree(),
            marker="^",
            color=st_dot_color0,
            markeredgecolor="k",
            markersize=8,
            markeredgewidth=0.5,
        )
        map_ax.text(stalons[i], stalats[i], name, fontsize=9, color=st_text_color0)

    if XC.rotation:
        dx = XC.rotation["x"][1] - XC.rotation["x"][0]
        dy = XC.rotation["y"][1] - XC.rotation["y"][0]
        dz = XC.rotation["z"][1] - XC.rotation["z"][0]
        plt.title(
            "Stations and Grid Nodes\ndx={:.1f} km, dy={:.1f} km, dz={:.1f} km\nDepth range: {:.1f} - {:.1f} km".format(
                dx, dy, dz, XC.rotation["z"][0], XC.rotation["z"][-1]
            ),
            fontsize=8,
        )
    else:
        dx = (
            gps2dist_azimuth(
                XC.grid_size["lats"][0],
                XC.grid_size["lons"][0],
                XC.grid_size["lats"][0],
                XC.grid_size["lons"][1],
            )[0]
            / 1000
        )
        dy = (
            gps2dist_azimuth(
                XC.grid_size["lats"][0],
                XC.grid_size["lons"][0],
                XC.grid_size["lats"][1],
                XC.grid_size["lons"][0],
            )[0]
            / 1000
        )
        dz = XC.grid_size["deps"][1] - XC.grid_size["deps"][0]
        plt.title(
            "Stations and Grid Nodes\ndx={:.2f} km, dy={:.2f} km, dz={:.2f} km\nDepth range {:.1f} - {:.1f} km".format(
                dx, dy, dz, XC.grid_size["deps"][0], XC.grid_size["deps"][-1]
            ),
            fontsize=8,
        )

    plt.show()

    ########################################
    ########################################

    return


def plot_locations(locs, XC):

    gs = gridspec.GridSpec(4, 1)
    fig = plt.figure(figsize=(6, 6))

    map_ax = fig.add_subplot(gs[:3, 0], projection=ccrs.PlateCarree())
    if XC.rotation:
        latlims = [XC._grid["LAT"].flatten().min(), XC._grid["LAT"].flatten().max()]
        lonlims = [XC._grid["LON"].flatten().min(), XC._grid["LON"].flatten().max()]
    else:
        latlims = [XC.grid_size["lats"][0], XC.grid_size["lats"][-1]]
        lonlims = [XC.grid_size["lons"][0], XC.grid_size["lons"][-1]]
    map_ax.set_extent(lonlims + latlims, crs=ccrs.PlateCarree())
    map_ax.add_feature(cfeature.LAND, color="silver")
    map_ax.add_feature(cfeature.OCEAN, color="lightblue")
    map_ax.add_feature(cfeature.COASTLINE)
    map_ax.add_feature(cfeature.BORDERS, linestyle=":")
    map_ax.add_feature(cfeature.LAKES, color="lightblue", alpha=0.5)
    map_ax.add_feature(cfeature.RIVERS, color="lightblue")

    # add lat/lon labels
    gl = map_ax.gridlines(
        crs=ccrs.PlateCarree(),
        draw_labels=True,
        linewidth=0.5,
        color="white",
        alpha=0.8,
        linestyle=":",
    )
    gl.xformatter = LongitudeFormatter(zero_direction_label=True)
    gl.yformatter = LatitudeFormatter()
    meridians = np.linspace(lonlims[0], lonlims[-1], 4)
    parallels = np.linspace(latlims[0], latlims[-1], 4)
    gl.xlocator = mticker.FixedLocator(meridians)
    gl.ylocator = mticker.FixedLocator(parallels)
    gl.bottom_labels = False
    gl.right_labels = False
    gl.xlabel_style = {"size": 6}
    gl.ylabel_style = {"size": 6}

    stalats = np.array([tr.stats.coordinates.latitude for tr in XC.traces])
    stalons = np.array([tr.stats.coordinates.longitude for tr in XC.traces])
    stanames = [tr.stats.station for tr in XC.traces]

    hist_ax = fig.add_subplot(gs[-1, 0])
    hist_ax.set_xlim(-1, len(XC.traces))
    hist_ax.set_title("Station Usage", fontsize=10)

    for i, name in enumerate(stanames):
        map_ax.plot(
            stalons[i],
            stalats[i],
            transform=ccrs.PlateCarree(),
            marker="^",
            color=st_dot_color0,
            markeredgecolor="k",
            markersize=8,
            markeredgewidth=0.5,
        )
        map_ax.text(stalons[i], stalats[i], name, fontsize=6, color=st_text_color0)

    if "event_list" in str(locs.__class__):
        map_ax.scatter(
            locs.get_lons(),
            locs.get_lats(),
            transform=ccrs.PlateCarree(),
            marker="o",
            s=18,
            color="red",
            edgecolors="k",
            linewidths=0.5,
        )
        stas = locs.get_stations()
        sta_list = dict(Counter(stas).most_common())
        for tr in XC.traces:
            if tr.id not in sta_list.keys():
                sta_list[tr.id] = 0
        plt.bar(sta_list.keys(), sta_list.values(), width=0.5, color="gray")

    elif "detections" in str(locs.__class__):
        # 1st plot noise
        h1 = map_ax.scatter(
            locs.noise.get_lons(),
            locs.noise.get_lats(),
            transform=ccrs.PlateCarree(),
            marker="o",
            s=8,
            color="lightsteelblue",
            edgecolors="k",
            linewidths=0.5,
        )
        # 2nd plot edge
        h2 = map_ax.scatter(
            locs.edge_clustered.get_lons(),
            locs.edge_clustered.get_lats(),
            transform=ccrs.PlateCarree(),
            marker="o",
            s=12,
            color="darkorange",
            edgecolors="k",
            linewidths=0.5,
        )
        # 2nd plot edge
        h3 = map_ax.scatter(
            locs.core_clustered.get_lons(),
            locs.core_clustered.get_lats(),
            transform=ccrs.PlateCarree(),
            marker="o",
            s=18,
            color="red",
            edgecolors="k",
            linewidths=0.5,
        )

        map_ax.legend(
            [h1, h2, h3],
            tuple(["noise", "edge", "core"]),
            loc="lower left",
            ncol=1,
            fontsize=8,
        )

        stas = locs.detections.get_stations()
        sta_list = dict(Counter(stas).most_common())
        for tr in XC.traces:
            if tr.id not in sta_list.keys():
                sta_list[tr.id] = 0

        b1 = hist_ax.bar(
            sta_list.keys(), sta_list.values(), width=0.5, color="lightsteelblue"
        )

        stas = locs.core_clustered.get_stations()
        sta_list = dict(Counter(stas).most_common())
        for tr in XC.traces:
            if tr.id not in sta_list.keys():
                sta_list[tr.id] = 0
        b2 = hist_ax.bar(sta_list.keys(), sta_list.values(), width=0.5, color="red")

        stas = locs.edge_clustered.get_stations()
        sta_list = dict(Counter(stas).most_common())
        for tr in XC.traces:
            if tr.id not in sta_list.keys():
                sta_list[tr.id] = 0
        b3 = hist_ax.bar(
            sta_list.keys(), sta_list.values(), width=0.5, color="darkorange"
        )

        hist_ax.legend(
            [b1, b3, b2],
            tuple(["noise", "edge", "core"]),
            loc="upper right",
            ncol=1,
            fontsize=6,
        )
        hist_ax.set_ylabel("counts")

    plt.xticks(rotation=45, fontsize=6, ha="right")

    plt.show()

    return
